# -*- coding: utf-8 -*-

from numpy import min as np_min, max as np_max, abs as np_abs, log10

import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.art3d as art3d
from matplotlib.image import NonUniformImage
import matplotlib

from SciDataTool.Functions.Plot.init_fig import init_fig
from SciDataTool.Functions.Plot import COLORS


def plot_3D(
    Xdata,
    Ydata,
    Zdata,
    colormap="RdBu_r",
    color_list=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    z_min=None,
    z_max=None,
    title="",
    xlabel="",
    ylabel="",
    zlabel="",
    xticks=None,
    yticks=None,
    fig=None,
    ax=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_logscale_z=False,
    is_disp_title=True,
    type_plot="stem",
    is_contour=False,
    save_path=None,
    is_show_fig=None,
    is_switch_axes=False,
    win_title=None,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
):
    """Plots a 3D graph ("stem", "surf" or "pcolor")

    Parameters
    ----------
    Xdata : ndarray
        array of x-axis values
    Ydata : ndarray
        array of y-axis values
    Zdata : ndarray
        array of z-axis values
    colormap : colormap object
        colormap prescribed by user
    x_min : float
        minimum value for the x-axis (no automated scaling in 3D)
    x_max : float
        maximum value for the x-axis (no automated scaling in 3D)
    y_min : float
        minimum value for the y-axis (no automated scaling in 3D)
    y_max : float
        maximum value for the y-axis (no automated scaling in 3D)
    z_min : float
        minimum value for the z-axis (no automated scaling in 3D)
    z_max : float
        maximum value for the z-axis (no automated scaling in 3D)
    title : str
        title of the graph
    xlabel : str
        label for the x-axis
    ylabel : str
        label for the y-axis
    zlabel : str
        label for the z-axis
    xticks : list
        list of ticks to use for the x-axis
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_logscale_z : bool
        boolean indicating if the z-axis must be set in logarithmic scale
    is_disp_title : bool
        boolean indicating if the title must be displayed
    type_plot : str
        type of 3D graph : "stem", "surf", "pcolor" or "scatter"
    is_contour : bool
        True to show contour line if type_plot = "pcolor"
    save_path : str
        full path including folder, name and extension of the file to save if save_path is not None
    is_show_fig : bool
        True to show figure after plot
    is_switch_axes : bool
        to switch x and y axes
    """

    # Set if figure must be shown if is_show_fig is None
    if is_show_fig is None:
        is_show_fig = True if fig is None else False

    # Set if figure is 3D
    if type_plot not in ["pcolor", "pcolormesh", "scatter"]:
        is_3d = True
    else:
        is_3d = False

    # Set figure if needed
    if fig is None and ax is None:
        (fig, ax, _, _) = init_fig(fig=None, shape="rectangle", is_3d=is_3d)

    if color_list is None:
        color_list = COLORS

    # Calculate z limits
    if z_min is None:
        z_min = np_min(Zdata)
    if z_max is None:
        z_max = np_max(Zdata)

    # Check logscale on z axis
    if is_logscale_z:
        Zdata = 10 * log10(np_abs(Zdata))
        clb_format = "%0.0f"
    else:
        clb_format = "%.4g"

    # Switch axes
    if is_switch_axes:
        Xdata, Ydata = Ydata, Xdata
        if len(Xdata.shape) > 1:
            Xdata = Xdata.T
        if len(Ydata.shape) > 1:
            Ydata = Ydata.T
        if len(Zdata.shape) > 1:
            Zdata = Zdata.T
        x_min, y_min = y_min, x_min
        x_max, y_max = y_max, x_max
        xlabel, ylabel = ylabel, xlabel
        xticks, yticks = yticks, xticks
        is_logscale_x, is_logscale_y = is_logscale_y, is_logscale_x

    # Plot
    if type_plot == "stem":
        cmap = matplotlib.cm.get_cmap(colormap)
        for xi, yi, zi in zip(Xdata, Ydata, Zdata):
            line = art3d.Line3D(
                *zip((xi, yi, 0), (xi, yi, zi)),
                linewidth=2.0,
                marker="o",
                markersize=3.0,
                markevery=(1, 1),
                color=cmap(0)
            )
            ax.add_line(line)
        ax.set_xlim3d(x_max, x_min)
        ax.set_ylim3d(y_min, y_max)
        ax.set_zlim3d(z_min, z_max)
        # set correct angle
        ax.view_init(elev=20.0, azim=45)
        ax.zaxis.set_rotate_label(False)
        ax.set_zlabel(zlabel, rotation=0)
        ax.xaxis.labelpad = 5
        ax.yaxis.labelpad = 5
        ax.zaxis.labelpad = 5
        if xticks is not None:
            ax.xaxis.set_ticks(xticks)
        if yticks is not None:
            ax.yaxis.set_ticks(yticks)
        # white background
        ax.xaxis.pane.fill = False
        ax.yaxis.pane.fill = False
        ax.zaxis.pane.fill = False
        ax.xaxis.pane.set_edgecolor("w")
        ax.yaxis.pane.set_edgecolor("w")
        ax.zaxis.pane.set_edgecolor("w")
        if is_logscale_z:
            ax.zscale("log")
    elif type_plot == "surf":
        ax.plot_surface(Xdata, Ydata, Zdata, cmap=colormap)
        ax.set_xlim3d(x_max, x_min)
        ax.set_ylim3d(y_min, y_max)
        ax.set_zlim3d(z_min, z_max)
        ax.zaxis.set_rotate_label(False)
        ax.set_zlabel(zlabel, rotation=0)
        ax.xaxis.labelpad = 5
        ax.yaxis.labelpad = 5
        ax.zaxis.labelpad = 5
        if xticks is not None:
            ax.xaxis.set_ticks(xticks)
        if yticks is not None:
            ax.yaxis.set_ticks(yticks)
        # white background
        ax.xaxis.pane.fill = False
        ax.yaxis.pane.fill = False
        ax.zaxis.pane.fill = False
        ax.xaxis.pane.set_edgecolor("w")
        ax.yaxis.pane.set_edgecolor("w")
        ax.zaxis.pane.set_edgecolor("w")
        if is_logscale_z:
            ax.zscale("log")
    elif type_plot == "pcolor":
        Zdata[Zdata < z_min] = z_min
        Zdata[Zdata > z_max] = z_max
        # Handle descending order axes (e.g. spectrogram of run-down in freq/rpm map)
        if Ydata[-1] < Ydata[0]:
            Ydata = Ydata[::-1]
            Zdata = Zdata[:, ::-1]
        if Xdata[-1] < Xdata[0]:
            Xdata = Xdata[::-1]
            Zdata = Zdata[::-1, :]
        im = NonUniformImage(
            ax,
            interpolation="bilinear",
            extent=(x_min, x_max, y_max, y_min),
            cmap=colormap,
            picker=10,
        )
        im.set_data(Xdata, Ydata, Zdata.T)
        ax.images.append(im)
        if is_contour:
            ax.contour(Xdata, Ydata, Zdata.T, colors="black", linewidths=0.8)
        clb = fig.colorbar(im, ax=ax, format=clb_format)
        clb.ax.set_title(zlabel, fontsize=font_size_legend, fontname=font_name)
        clb.ax.tick_params(labelsize=font_size_legend)
        for l in clb.ax.yaxis.get_ticklabels():
            l.set_family(font_name)
        if xticks is not None:
            ax.xaxis.set_ticks(xticks)
        if yticks is not None:
            ax.yaxis.set_ticks(yticks)
        ax.set_xlim([x_min, x_max])
        ax.set_ylim([y_min, y_max])
    elif type_plot == "pcolormesh":
        c = ax.pcolormesh(
            Xdata, Ydata, Zdata, cmap=colormap, shading="gouraud", antialiased=True
        )
        clb = fig.colorbar(c, ax=ax)
        clb.ax.set_title(zlabel, fontsize=font_size_legend, fontname=font_name)
        clb.ax.tick_params(labelsize=font_size_legend)
        for l in clb.ax.yaxis.get_ticklabels():
            l.set_family(font_name)
        if xticks is not None:
            ax.xaxis.set_ticks(xticks)
        if yticks is not None:
            ax.yaxis.set_ticks(yticks)
        ax.set_xlim([x_min, x_max])
        ax.set_ylim([y_min, y_max])
    elif type_plot == "scatter":
        c = ax.scatter(
            Xdata,
            Ydata,
            # s=10,
            c=Zdata,
            marker=".",
            cmap=colormap,
            vmin=z_min,
            vmax=z_max,
            picker=True,
            pickradius=5,
        )
        clb = fig.colorbar(c, ax=ax)
        clb.ax.set_title(zlabel, fontsize=font_size_legend, fontname=font_name)
        clb.ax.tick_params(labelsize=font_size_legend)
        for l in clb.ax.yaxis.get_ticklabels():
            l.set_family(font_name)
        if xticks is not None:
            ax.xaxis.set_ticks(xticks)
        if yticks is not None:
            ax.yaxis.set_ticks(yticks)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    if is_logscale_x:
        ax.xscale("log")

    if is_logscale_y:
        ax.yscale("log")

    if is_disp_title:
        ax.set_title(title)

    if is_3d:
        for item in (
            [ax.xaxis.label, ax.yaxis.label, ax.zaxis.label]
            + ax.get_xticklabels()
            + ax.get_yticklabels()
            + ax.get_zticklabels()
        ):
            item.set_fontsize(font_size_label)
    else:
        for item in (
            [ax.xaxis.label, ax.yaxis.label]
            + ax.get_xticklabels()
            + ax.get_yticklabels()
        ):
            item.set_fontsize(font_size_label)
            item.set_fontname(font_name)
    ax.title.set_fontsize(font_size_title)
    ax.title.set_fontname(font_name)

    if save_path is not None:
        save_path = save_path.replace("\\", "/")
        fig.savefig(save_path)
        plt.close()

    if is_show_fig:
        fig.show()

    if win_title:
        manager = plt.get_current_fig_manager()
        if manager is not None:
            manager.set_window_title(win_title)



# -*- coding: utf-8 -*-
from numpy import take


def extract_slices_fft(self, values, axes_list):
    """Returns the values of the field (with symmetries and transformations).
    Parameters
    ----------
    self: Data
        a Data object
    values: ndarray
        array of the field
    axes_list: list
        a list of RequestedAxis objects
    Returns
    -------
    values: ndarray
        values of the field
    """

    # Extract the slices of the field
    for index, axis in enumerate(self.axes):
        is_match = False
        for axis_requested in axes_list:
            if axis.name == axis_requested.corr_name:
                is_match = True
                if (
                    axis_requested.indices is not None
                    and axis_requested.transform == "fft"
                ):
                    values = take(values, axis_requested.indices, axis=index)
        if not is_match:  # Axis was not specified -> take slice at the first value
            values = take(values, [0], axis=index)
    return values



# -*- coding: utf-8 -*-
from SciDataTool.Functions import UnitError
from SciDataTool.Functions.fft_functions import (
    comp_fft_freqs,
    comp_fft_time,
    comp_nthoctave_axis,
)
from numpy import (
    pi,
    log10,
    sqrt,
    square,
    column_stack,
    exp,
    real,
    imag,
    cos,
    sin,
    ndarray,
    nan,
    isnan,
    abs as np_abs,
    sum as np_sum,
    angle as np_angle,
    where,
    argwhere,
    array,
    take,
    zeros,
    floor,
)

# List of the unit symbols, their normalizing value and their dimensions "MLTTempAngleCurrent"
unit_symbols = [
    ("dimless", 1.0, (0, 0, 0, 0, 0, 0)),  # dimensionless
    ("degC", 1.0, (0, 0, 0, 1, 0, 0)),  # degree Celsius
    ("rad", 1.0, (0, 0, 0, 0, 1, 0)),  # radians
    ("°", pi / 180, (0, 0, 0, 0, 1, 0)),  # degree (angle)
    ("rpm", 2 * pi / 60, (0, 0, -1, 0, 1, 0)),  # rotation per minute
    ("min", 60.0, (0, 0, 1, 0, 0, 0)),  # minute
    ("Ohm", 1.0, (1, 2, -3, 0, 0, -2)),  # Ohm
    ("At", 1.0, (0, 0, 0, 0, 0, 1)),  # Ampere-tour
    ("Wb", 1.0, (1, 2, -2, 0, 0, -1)),  # Weber
    ("Mx", 1.0e-8, (1, 2, -2, 0, 0, -1)),  # Maxwell
    ("Hz", 1.0, (0, 0, -1, 0, 0, 0)),  # Herz
    ("Pa", 1.0, (1, -1, -2, 0, 0, 0)),  # Pascal
    ("g", 1.0e-3, (0, 1, 0, 0, 0, 0)),  # gram
    ("s", 1.0, (0, 0, 1, 0, 0, 0)),  # second
    ("h", 3600.0, (0, 0, 1, 0, 0, 0)),  # hour
    #    ("K",       1.0, (0,0,0,1,0,0)),   # Kelvin
    ("A", 1.0, (0, 0, 0, 0, 0, 1)),  # Ampere
    ("J", 1.0, (1, 2, -2, 0, 0, 0)),  # Joule
    ("W", 1.0, (1, 2, -3, 0, 0, 0)),  # Watt
    ("N", 1.0, (1, 1, -2, 0, 0, 0)),  # Newton
    ("C", 1.0, (0, 0, 1, 0, 0, 1)),  # Coulomb
    ("T", 1.0, (1, 0, -2, 0, 0, -1)),  # Tesla
    ("G", 1.0e-4, (1, 0, -2, 0, 0, -1)),  # Gauss
    ("V", 1.0, (1, 2, -3, 0, 0, -1)),  # Volt
    ("F", 1.0, (-1, -2, 4, 0, 0, 2)),  # Farrad
    ("H", 1.0, (1, -2, -2, 0, 0, -2)),  # Henry
    ("m", 1.0, (1, 0, 0, 0, 0, 0)),  # meter
]
# Dictionnary of the prefixes and their values
unit_prefixes = {
    "Y": 1e24,
    "Z": 1e21,
    "E": 1e18,
    "P": 1e15,
    "T": 1e12,
    "G": 1e9,
    "M": 1e6,
    "k": 1e3,
    "h": 1e2,
    "da": 1e1,
    "": 1.0,
    "d": 1e-1,
    "c": 1e-2,
    "m": 1e-3,
    "µ": 1e-6,  # ('MICRO SIGN' U+00B5)
    "u": 1e-6,
    "μ": 1e-6,  # ('GREEK SMALL LETTER MU' U+03BC)
    "n": 1e-9,
    "p": 1e-12,
    "f": 1e-15,
    "a": 1e-18,
    "z": 1e-21,
    "y": 1e-24,
}


def get_dim_prefix(unit_str):
    if unit_str == "":
        unit_str = "dimless"
    p = 1  # power of the unit
    dim = None
    for key in unit_symbols:
        if key[0] in unit_str:
            if unit_str.rsplit(key[0], 1)[1].isdigit():
                p = int(unit_str.rsplit(key[0], 1)[1])
            dim = [p * d for d in key[2]]
            prefix_str = unit_str.rsplit(key[0], 1)[0]
            if prefix_str in unit_prefixes.keys():
                prefix = (unit_prefixes.get(prefix_str) * key[1]) ** p
            else:
                raise UnitError("ERROR: Prefix " + prefix_str + " unknown")
            break
    if not dim:
        raise UnitError("ERROR: Unit " + unit_str + " unknown")
    return (dim, prefix)


def convert(values, unit1, unit2):
    """Converts values from unit1 to unit2
    Parameters
    ----------
    values: ndarray
        Values of the field to convert
    unit1: str
        start unit
    unit2: str
        final unit
    Returns
    -------
    ndarray of the converted field
    """
    # Critical band to frequency conversion (for acoustic loudness)
    # Traunmüller, H. (1990). "Analytical expressions for the tonotopic sensory scale".
    # The Journal of the Acoustical Society of America. 88: 97. doi:10.1121/1.399849
    if unit1 == "Bark" and unit2 == "Hz":
        return 1960 * (values + 0.53) / (26.28 - values)
    elif unit1 == "Hz" and unit2 == "Bark":
        return (26.81 * values / (1960 + values)) - 0.53
    # Generic conversion
    else:
        unit1_save = unit1
        unit2_save = unit2
        # Format the strings
        unit1 = unit1.replace("*", "").replace(" ", "").replace("^", "")
        unit2 = unit2.replace("*", "").replace(" ", "").replace("^", "")
        # Unit1 parsing
        if "/" in unit1:
            dim1_denom, prefix1_denom = get_dim_prefix(unit1.split("/")[1])
            unit1 = unit1.split("/")[0]
        else:
            dim1_denom = [0, 0, 0, 0, 0, 0]
            prefix1_denom = 1.0
        dim1_num, prefix1_num = get_dim_prefix(unit1)
        # Unit2 parsing
        if "/" in unit2:
            dim2_denom, prefix2_denom = get_dim_prefix(unit2.split("/")[1])
            unit2 = unit2.split("/")[0]
        else:
            dim2_denom = [0, 0, 0, 0, 0, 0]
            prefix2_denom = 1.0
        dim2_num, prefix2_num = get_dim_prefix(unit2)
        # Check compatibility
        dim1 = [i - j for i, j in zip(dim1_num, dim1_denom)]
        dim2 = [i - j for i, j in zip(dim2_num, dim2_denom)]
        if dim1 != dim2:
            raise UnitError(
                "ERROR: Units " + unit1_save + " and " + unit2_save + " do not match"
            )
        else:
            return (
                values * (prefix1_num / prefix1_denom) / (prefix2_num / prefix2_denom)
            )


def to_dB(values, unit, ref_value=1.0):
    """Converts values into dB normalized with ref_value

    Parameters
    ----------
    values: ndarray
        Values of the field to convert
    unit: str
        Unit
    ref_value: float
        Reference value
    Returns
    -------
    ndarray of the converted field
    """
    if ref_value != 1:
        if isinstance(values, ndarray):
            values[values < ref_value] = ref_value
        else:
            if values < ref_value:
                values = ref_value
    mask = values != 0
    try:
        convert(values, unit, "W")
        return 10.0 * where(mask, log10(values / ref_value, where=mask), 0)
    except:
        return 20.0 * where(mask, log10(values / ref_value, where=mask), 0)


def to_dBA(values, freqs, unit, ref_value=1.0):
    """Converts values into dBA (requires frequency vector)

    Parameters
    ----------
    values: array
        Values of the field to convert (must be 1D)
    freqs: array
        Frequency vector
    unit: str
        Unit
    ref_value: float
        Reference value
    Returns
    -------
    ndarray of the converted field
    """
    values = to_dB(values, unit, ref_value)
    return dB_to_dBA(values, freqs)


def dB_to_dBA(values, freqs, noct=None):
    """Converts values from dB into dBA (requires frequency vector)

    Parameters
    ----------
    values: array
        Values of the field to convert (must be 1D)
    freqs: array
        Frequency vector
    Returns
    -------
    ndarray of the converted field
    """
    if noct is not None:
        freqs_ref = [
            12.5,
            16,
            20,
            25,
            31.5,
            40,
            50,
            63,
            80,
            100,
            125,
            160,
            200,
            250,
            315,
            400,
            500,
            630,
            800,
            1000,
            1250,
            1600,
            2000,
            2500,
            3150,
            4000,
            5000,
            6300,
            8000,
            10000,
            12500,
            16000,
            20000,
        ]
        Aweight = [
            -63.4,
            -56.7,
            -50.5,
            -44.7,
            -39.4,
            -34.6,
            -30.2,
            -26.2,
            -22.5,
            -19.1,
            -16.1,
            -13.4,
            -10.9,
            -8.6,
            -6.6,
            -4.8,
            -3.2,
            -1.9,
            -0.8,
            0,
            0.6,
            1,
            1.2,
            1.3,
            1.2,
            1,
            0.5,
            -0.1,
            -1.1,
            -2.5,
            -4.3,
            -6.6,
            -9.3,
        ]
        mask = argwhere(freqs in freqs_ref)
        Aweight = take(Aweight, mask)
    else:
        freq2 = square(freqs)
        freq2[freq2 == 0] = nan
        RA = (
            12200.0 ** 2
            * freq2 ** 2
            / (
                (freq2 + 20.6 ** 2)
                * sqrt((freq2 + 107.7 ** 2) * (freq2 + 737.0 ** 2))
                * (freq2 + 12200.0 ** 2)
            )
        )
        Aweight = 2.0 + 20.0 * log10(RA)
        Aweight[isnan(Aweight)] = -100  # replacing NaN by -100 dB
        Aweight[
            values <= 0
        ] = 0  # avoiding to increase dB in dBA at frequencies where noise is already null
    try:
        values += Aweight
        return values
    except:
        raise UnitError("ERROR: dBA conversion only available for 1D fft")


def to_noct(values, freqs, noct=3):
    """Converts values into 1/nth octave band (requires frequency vector)

    Parameters
    ----------
    values: array
        Values of the field to convert (must be 1D)
    freqs: array
        Frequency vector
    noct: int
        Requested octave band
    Returns
    -------
    ndarray of the converted field
    """
    # Compute the 1/n octave axis
    f_oct = comp_nthoctave_axis(noct, freqs[0], freqs[-1])
    # Compute sum over each interval
    freqbds = [f / (2 ** (1.0 / (2.0 * noct))) for f in f_oct]
    freqbds.append(freqbds[-1])
    # freqbds = [0] + f_oct
    values_oct = []
    N = zeros(len(freqbds) - 1)
    for i in range(len(freqbds) - 1):
        f1 = freqbds[i]
        f2 = freqbds[i + 1]
        indices = argwhere((freqs >= f1) & (freqs <= f2))
        N[i] = len(indices)
        if len(indices) == 0:
            values_oct.append(0)
        else:
            values2 = take(values, indices)
            if np_sum(values2) == 0:
                values_oct.append(0)
            else:
                values_oct.append(sqrt(np_sum(values2 ** 2)))
    values = array(values_oct)
    # Apply bin correction factor for when centre frequencies are not integer
    # multiples of the bandwicth (BW)
    df_fft = freqs[1] - freqs[0]
    BW = zeros(len(freqbds) - 1)
    for i in range(len(freqbds) - 1):
        BW[i] = freqbds[i + 1] - freqbds[i]
    for i in range(len(freqbds) - 1):
        if (N[i] != 0) and (floor(BW[i] / df_fft) <= N[i]):
            values[i] = values[i] * (BW[i] / (df_fft * N[i])) ** 0.5
    return [values, array(f_oct)]


def xyz_to_rphiz(values):
    """Converts axis values from cartesian coordinates into cylindrical coordinates

    Parameters
    ----------
    values: array
        Values of the axis to convert (Nx3)
    Returns
    -------
    ndarray of the axis (Nx3)
    """

    x = values[:, 0]
    y = values[:, 1]
    z = values[:, 2]

    affixe = x + 1j * y
    r = np_abs(affixe)
    phi = (np_angle(affixe) + 2 * pi) % (2 * pi)

    return column_stack((r, phi, z))


def rphiz_to_xyz(values):
    """Converts axis values from cylindrical coordinates into cartesian coordinates

    Parameters
    ----------
    values: array
        Values of the axis to convert (Nx3)
    Returns
    -------
    ndarray of the axis (Nx3)
    """

    r = values[:, 0]
    phi = values[:, 1]
    if len(values.shape) > 2:
        z = values[:, 2]

    affixe = r * exp(1j * phi)
    x = real(affixe)
    y = imag(affixe)

    if len(values.shape) > 2:
        return column_stack((x, y, z))
    else:
        return column_stack((x, y))


def xyz_to_rphiz_field(values, phi):
    """Converts field values from cartesian coordinates into cylindrical coordinates

    Parameters
    ----------
    values: array
        Values of the field to convert (Nx3)
    phi: array
        Values of the angle axis (N)
    Returns
    -------
    ndarray of the field (Nx3)
    """

    field_x = values[:, 0]
    field_y = values[:, 1]
    field_z = values[:, 2]

    cos_phi = cos(phi)
    sin_phi = sin(phi)

    field_r = cos_phi * field_x + sin_phi * field_y
    field_phi = -sin_phi * field_x + cos_phi * field_y

    return column_stack((field_r, field_phi, field_z))


def rphiz_to_xyz_field(values, phi):
    """Converts field values from cylindrical coordinates into cartesian coordinates

    Parameters
    ----------
    values: array
        Values of the field to convert (Nx3)
    phi: array
        Values of the angle axis (N)
    Returns
    -------
    ndarray of the field (Nx3)
    """

    field_r = values[:, 0]
    field_phi = values[:, 1]
    field_z = values[:, 2]

    cos_phi = cos(phi)
    sin_phi = sin(phi)

    field_x = cos_phi * field_r - sin_phi * field_phi
    field_y = sin_phi * field_r + cos_phi * field_phi

    return column_stack((field_x, field_y, field_z))


def time_to_freqs(values, is_real):
    is_time = True
    return comp_fft_freqs(values, is_time, is_real)


def freqs_to_time(values, is_real):
    is_angle = False
    return comp_fft_time(values, is_angle, is_real)


def angle_to_wavenumber(values, is_real):
    is_time = False
    return comp_fft_freqs(values, is_time, is_real)


def wavenumber_to_angle(values, is_real):
    is_angle = True
    return comp_fft_time(values, is_angle, is_real)


def xy_to_rphi(x, y):
    affixe = x + 1j * y
    r = np_abs(affixe)
    phi = (np_angle(affixe) + 2 * pi) % (2 * pi)

    return (r, phi)


def rphi_to_xy(r, phi):
    affixe = r * exp(1j * phi)
    x = real(affixe)
    y = imag(affixe)

    return (x, y)


def cart2pol(field_x, field_y, phi):
    cos_phi = cos(phi)
    sin_phi = sin(phi)

    field_r = cos_phi * field_x + sin_phi * field_y
    field_phi = -sin_phi * field_x + cos_phi * field_y

    return (field_r, field_phi)


def pol2cart(field_r, field_phi, phi):
    cos_phi = cos(phi)
    sin_phi = sin(phi)

    field_x = cos_phi * field_r - sin_phi * field_phi
    field_y = sin_phi * field_r + cos_phi * field_phi

    return (field_x, field_y)



from json import load as jload
from os.path import isfile, isdir
from re import match


def load_json(file_path):
    """Load a json file

    Parameters
    ----------
    file_path: str
        path to the file to load

    Returns
    -------
    file_path: str
        edited path to the file to load
    json_data: json decoded data type
        data of the json file
    """
    if isdir(file_path):
        i = max(file_path.rfind("\\"), file_path.rfind("/"))
        if i != -1:
            file_path += file_path[i:] + ".json"
        else:
            file_path += "/" + file_path + ".json"

    # The file_name must end with .json
    elif not match(".*\.json", file_path):
        file_path += ".json"  # If it doesn't, we add .json at the end

    # The file (and the folder) should exist
    if not isfile(file_path):
        raise LoadMissingFileError(str(file_path) + " doesn't exist")

    # Get the data dictionary
    with open(file_path, "r") as load_file:
        json_data = jload(load_file)

    return file_path, json_data


class LoadMissingFileError(Exception):
    """ """

    pass



# -*- coding: utf-8 -*-
from SciDataTool import Data1D
from SciDataTool.Functions import AxisError


def freq_to_time(self):
    """Performs the inverse Fourier Transform and stores the resulting field in a DataTime object.
    Parameters
    ----------
    self : DataFreq
        a DataFreq object
    Returns
    -------
    a DataTime object
    """

    # Dynamic import to avoid loop
    module = __import__("SciDataTool.Classes.DataTime", fromlist=["DataTime"])
    DataTime = getattr(module, "DataTime")
    module = __import__("SciDataTool.Classes.DataPattern", fromlist=["DataPattern"])
    DataPattern = getattr(module, "DataPattern")

    axes_str = []
    for i, axis in enumerate(self.axes):
        if axis.is_components:
            axis_str = axis.name + str(list(range(len(axis.values))))
        elif axis.name == "freqs":
            axis_str = "time[smallestperiod]"
        elif axis.name == "wavenumber":
            axis_str = "angle[smallestperiod]"
        elif isinstance(axis, DataPattern):
            axis_str = axis.name + "[pattern]"
        else:
            axis_str = axis.name + "[smallestperiod]"
        axes_str.append(axis_str)
    if axes_str == [axis.name for axis in self.axes]:
        raise AxisError(
            "ERROR: No available axis is compatible with fft (should be time or angle)"
        )
    else:
        results = self.get_along(*axes_str)
        values = results.pop(self.symbol)
        Axes = []
        for axis in self.axes:
            if axis.name == "freqs":
                axis_new = Data1D(
                    name="time",
                    is_components=False,
                    values=results["time"],
                    unit="s",
                    symmetries=axis.symmetries.copy(),
                    normalizations=axis.normalizations.copy(),
                )
            elif axis.name == "wavenumber":
                axis_new = Data1D(
                    name="angle",
                    is_components=False,
                    values=results["angle"],
                    unit="rad",
                    symmetries=axis.symmetries.copy(),
                    normalizations=axis.normalizations.copy(),
                )
            else:
                axis_new = axis.copy()
            Axes.append(axis_new)

        return DataTime(
            name=self.name,
            unit=self.unit,
            symbol=self.symbol,
            axes=Axes,
            values=values,
            is_real=self.is_real,
            normalizations=self.normalizations.copy(),
        )



# -*- coding: utf-8 -*-
from SciDataTool.Functions import NormError
from SciDataTool.Functions.conversions import convert
from numpy import angle as np_angle


def get_phase_along(self, *args, unit="SI", is_norm=False, axis_data=[]):
    """Returns the ndarray of the magnitude of the FT, using conversions and symmetries if needed.
    Parameters
    ----------
    self: Data
        a Data object
    *args: list of strings
        List of axes requested by the user, their units and values (optional)
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    axis_data: list
        list of ndarray corresponding to user-input data
    Returns
    -------
    list of 1Darray of axes values, ndarray of magnitude values
    """
    if len(args) == 1 and type(args[0]) == tuple:
        args = args[0]  # if called from another script with *args
    return_dict = self.get_along(args, axis_data=axis_data)
    values = return_dict[self.symbol]
    # Compute magnitude
    values = np_angle(values)
    # Convert into right unit (apart because of degree conversion)
    if unit == self.unit or unit == "SI":
        if is_norm:
            try:
                values = values / self.normalizations.get("ref")
            except:
                raise NormError(
                    "ERROR: Reference value not specified for normalization"
                )
    elif unit == "°":
        values = convert(values, "rad", "°")
    elif unit in self.normalizations:
        values = values / self.normalizations.get(unit)
    else:
        values = convert(values, self.unit, unit)
    return_dict[self.symbol] = values
    return return_dict



# -*- coding: utf-8 -*-
from json import dump
from os.path import join, basename, isdir


def fix_file_name(save_path, obj):
    if isdir(save_path) or not save_path:
        file_path = join(save_path, type(obj).__name__ + ".json")
    elif ".json" != basename(save_path)[-5:]:
        file_path = save_path + ".json"
    else:
        file_path = save_path
    return file_path


def is_json_serializable(obj):
    if isinstance(obj, (bool, float, int, str)):
        return True
    else:
        return False


def has_as_dict(obj):
    """Check if object has 'as_dict' method."""
    return hasattr(obj, "as_dict") and callable(getattr(obj, "as_dict", None))


def build_data(obj):
    """
    Build a json serializable data structure of lists, dicts and pyleecan objects.
    Data that can not be serialized will be set to None. Tuples will also be None.
    Parameters
    ----------
    obj :
        An object to serialize

    Returns
    -------
    data :
        A serializable data structure
    """
    # lists
    if isinstance(obj, list):
        data = []
        for elem in obj:
            data.append(build_data(elem))
        return data
    # dicts
    if isinstance(obj, dict):
        data = {}
        for key in obj:
            data[key] = build_data(obj[key])
        return data
    # tuples (excluded)
    if isinstance(obj, tuple):
        return None
    # pyleecan classes, i.e. instances with as_dict method
    if has_as_dict(obj):
        return obj.as_dict()
    #
    if is_json_serializable(obj):
        return obj
    else:
        return None


def save_data(obj, save_path=""):
    """Save the object to the save_path
    Parameters
    ----------
    self :
        A pyleecan object
    save_path: str
        path to the folder to save the object
    """
    # correct file name if needed
    file_path = fix_file_name(save_path, obj)
    # save
    obj = build_data(obj)
    with open(file_path, "w") as json_file:
        dump(obj, json_file, sort_keys=True, indent=4, separators=(",", ": "))


def save(self, save_path=""):
    """Save the object to the save_path
    Parameters
    ----------
    self :
        A pyleecan object
    save_path: str
        path to the folder to save the object
    """
    save_data(self, save_path=save_path)



# -*- coding: utf-8 -*-
from SciDataTool.Classes.Data import Data
from SciDataTool.Classes.Data1D import Data1D
from SciDataTool.Classes.DataFreq import DataFreq
from SciDataTool.Classes.DataTime import DataTime
from SciDataTool.Classes.DataLinspace import DataLinspace
from SciDataTool.Classes.DataND import DataND
from SciDataTool.Classes.RequestedAxis import RequestedAxis
from SciDataTool.Classes.VectorField import VectorField
from SciDataTool.Classes.DataPattern import DataPattern

__version__ = "1.4.2"



# -*- coding: utf-8 -*-
from numpy import (
    argsort,
    negative,
    meshgrid,
)


def get_harmonics(
    self, N_harm, *args, unit="SI", is_norm=False, axis_data=None, is_flat=False
):
    """Returns the complex Fourier Transform of the field, using conversions and symmetries if needed.
    Parameters
    ----------
    self: Data
        a Data object
    N_harm: int
        Number of largest harmonics to be extracted
    args: list
        Axes names, ranges and units
    unit: str
        Unit demanded by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    is_flat: bool
        Boolean if the output data remains flattened (for 2D cases)
    Returns
    -------
    list of 1Darray of axes values, ndarray of magnitude of FT
    """
    if len(args) == 1 and type(args[0]) == tuple:
        args = args[0]  # if called from another script with *args
    return_dict = self.get_magnitude_along(
        args, unit=unit, is_norm=is_norm, axis_data=axis_data
    )
    values = return_dict[self.symbol]

    # 2D case
    if "freqs" in return_dict and "wavenumber" in return_dict:
        r = return_dict["wavenumber"]
        f = return_dict["freqs"]
        # Flatten the data
        values_flat = values.flatten()
        R, F = meshgrid(r, f)
        f = F.flatten()
        r = R.flatten()
        # Get the N_harm largest peaks
        indices = argsort(negative(values_flat))
        indices = indices[:N_harm]
        values = values_flat[indices]
        f = f[indices]
        r = r[indices]
        if len(values.shape) == 2 and not is_flat:
            f.reshape((N_harm, N_harm))
            r.reshape((N_harm, N_harm))
            values.reshape((N_harm, N_harm))
        return_dict["freqs"] = f
        return_dict["wavenumber"] = r
        return_dict[self.symbol] = values

    # 1D cases
    elif "freqs" in return_dict:
        f = return_dict["freqs"]
        indices = argsort(negative(values))
        indices = indices[:N_harm]
        f = f[indices]
        values = values[indices]
        return_dict["freqs"] = f
        return_dict[self.symbol] = values
    elif "wavenumber" in return_dict:
        r = return_dict["wavenumber"]
        indices = argsort(negative(values))
        indices = indices[:N_harm]
        r = r[indices]
        values = values[indices]
        return_dict["wavenumber"] = r
        return_dict[self.symbol] = values

    return return_dict



# -*- coding: utf-8 -*-
from SciDataTool import Data1D
from SciDataTool.Functions import AxisError, axes_dict, rev_axes_dict


def get_data_along(self, *args, unit="SI", is_norm=False, axis_data=[]):
    """Returns the sliced or interpolated version of the data, using conversions and symmetries if needed.
    Parameters
    ----------
    self: Data
        a Data object
    *args: list of strings
        List of axes requested by the user, their units and values (optional)
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    axis_data: list
        list of ndarray corresponding to user-input data
    Returns
    -------
    a DataND object
    """

    # Dynamic import to avoid loop
    module = __import__("SciDataTool.Classes.DataND", fromlist=["DataND"])
    DataND = getattr(module, "DataND")

    results = self.get_along(*args)
    values = results.pop(self.symbol)
    del results["axes_dict_other"]
    del results["axes_list"]
    Axes = []
    for axis_name in results.keys():
        if len(results[axis_name]) > 1:
            for axis in self.axes:
                if axis.name == axis_name:
                    name = axis.name
                    is_components = axis.is_components
                    axis_values = results[axis_name]
                    unit = axis.unit
                elif axis_name in axes_dict:
                    if axes_dict[axis_name][0] == axis.name:
                        name = axis_name
                        is_components = axis.is_components
                        axis_values = results[axis_name]
                        unit = axes_dict[axis_name][2]
                elif axis_name in rev_axes_dict:
                    if rev_axes_dict[axis_name][0] == axis.name:
                        name = axis_name
                        is_components = axis.is_components
                        axis_values = results[axis_name]
                        unit = rev_axes_dict[axis_name][2]
            Axes.append(
                Data1D(
                    name=name,
                    unit=unit,
                    values=axis_values,
                    is_components=is_components,
                )
            )
    return DataND(
        name=self.name,
        unit=self.unit,
        symbol=self.symbol,
        axes=Axes,
        values=values,
        is_real=self.is_real,
    )



# -*- coding: utf-8 -*-

from numpy import log10, abs as np_abs, max as np_max, NaN, zeros_like
import matplotlib.pyplot as plt

from SciDataTool.Functions.Plot.init_fig import init_fig


def plot_4D(
    Xdata,
    Ydata,
    Zdata,
    Sdata=None,
    is_same_size=False,
    colormap="RdBu_r",
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    z_min=None,
    z_max=None,
    title="",
    xlabel="",
    ylabel="",
    zlabel="",
    xticks=None,
    yticks=None,
    xticklabels=None,
    yticklabels=None,
    annotations=None,
    fig=None,
    ax=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_logscale_z=False,
    is_disp_title=True,
    type_plot="scatter",
    save_path=None,
    is_show_fig=None,
    is_switch_axes=False,
    win_title=None,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
    is_grid=False,
):
    """Plots a 4D graph

    Parameters
    ----------
    Xdata : ndarray
        array of x-axis values
    Ydata : ndarray
        array of y-axis values
    Zdata : ndarray
        array of z-axis values
    Sdata : ndarray
        array of 4th axis values
    is_same_size : bool
        in scatter plot, all squares are the same size
    colormap : colormap object
        colormap prescribed by user
    x_min : float
        minimum value for the x-axis (no automated scaling in 3D)
    x_max : float
        maximum value for the x-axis (no automated scaling in 3D)
    y_min : float
        minimum value for the y-axis (no automated scaling in 3D)
    y_max : float
        maximum value for the y-axis (no automated scaling in 3D)
    z_min : float
        minimum value for the z-axis (no automated scaling in 3D)
    z_max : float
        maximum value for the z-axis (no automated scaling in 3D)
    title : str
        title of the graph
    xlabel : str
        label for the x-axis
    ylabel : str
        label for the y-axis
    zlabel : str
        label for the z-axis
    xticks : list
        list of ticks to use for the x-axis
    yticks: list
        list of ticks to use for the y-axis
    xticklabels : list
        list of tick labels to use for the x-axis
    yticklabels : list
        list of tick labels to use for the x-axis
    annotations : list
        list of annotations to apply to data
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_logscale_z : bool
        boolean indicating if the z-axis must be set in logarithmic scale
    is_disp_title : bool
        boolean indicating if the title must be displayed
    type : str
        type of 3D graph : "stem", "surf", "pcolor" or "scatter"
    save_path : str
        full path including folder, name and extension of the file to save if save_path is not None
    is_show_fig : bool
        True to show figure after plot
    is_switch_axes : bool
        to switch x and y axes
    """

    # Set figure/subplot
    if is_show_fig is None:
        is_show_fig = True if fig is None else False

    # Set if figure is 3D
    if type_plot != "scatter":
        is_3d = True
    else:
        is_3d = False

    # Set figure if needed
    if fig is None and ax is None:
        (fig, ax, _, _) = init_fig(fig=None, shape="rectangle", is_3d=is_3d)

    # Check logscale on z axis
    if is_logscale_z:
        Zdata = 10 * log10(np_abs(Zdata))
        clb_format = "%0.0f"
    else:
        clb_format = "%.4g"

    # Calculate z limits
    if z_max is None:
        z_max = np_max(Zdata)
    if z_min is None:
        if is_logscale_z:
            z_min = 0
        else:
            z_min = z_max / 1e4

    Zdata[Zdata < z_min] = NaN

    if is_same_size:
        Sdata = zeros_like(Zdata)
        Sdata[Zdata > z_max / 1e4] = 40
    elif Sdata is None:
        Sdata = 500 * Zdata / z_max

    # Switch axes
    if is_switch_axes:
        Xdata, Ydata = Ydata, Xdata
        if len(Xdata.shape) > 1:
            Xdata = Xdata.T
        if len(Ydata.shape) > 1:
            Ydata = Ydata.T
        if len(Zdata.shape) > 1:
            Zdata = Zdata.T
        x_min, y_min = y_min, x_min
        x_max, y_max = y_max, x_max
        xlabel, ylabel = ylabel, xlabel
        xticks, yticks = yticks, xticks
        xticklabels, yticklabels = yticklabels, xticklabels
        is_logscale_x, is_logscale_y = is_logscale_y, is_logscale_x

    # Plot
    if type_plot == "scatter":
        c = ax.scatter(
            Xdata,
            Ydata,
            c=Zdata,
            s=Sdata,
            marker="s",
            cmap=colormap,
            vmin=z_min,
            vmax=z_max,
            picker=True,
            pickradius=5,
        )
        clb = fig.colorbar(c, ax=ax, format=clb_format)
        clb.ax.set_title(zlabel, fontsize=font_size_legend, fontname=font_name)
        clb.ax.tick_params(labelsize=font_size_legend)
        for l in clb.ax.yaxis.get_ticklabels():
            l.set_family(font_name)
        if xticks is not None:
            ax.xaxis.set_ticks(xticks)
        if xticklabels is not None:
            ax.set_xticklabels(xticklabels, rotation=90)
        if yticks is not None:
            ax.yaxis.set_ticks(yticks)
        if yticklabels is not None:
            ax.set_yticklabels(yticklabels)
        if annotations is not None:
            for i, txt in enumerate(annotations):
                if Zdata[i] > z_max * 0.01:
                    ax.annotate(
                        str(txt) + " [Hz]",
                        (Xdata[i], Ydata[i]),
                        rotation=45,
                        family=font_name,
                    )

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xlim([x_min, x_max])
    ax.set_ylim([y_min, y_max])

    if is_logscale_x:
        ax.xscale("log")

    if is_logscale_y:
        ax.yscale("log")

    if is_disp_title:
        ax.set_title(title)

    if is_3d:
        for item in (
            [ax.xaxis.label, ax.yaxis.label, ax.zaxis.label]
            + ax.get_xticklabels()
            + ax.get_yticklabels()
            + ax.get_zticklabels()
        ):
            item.set_fontsize(font_size_label)
    else:
        for item in (
            [ax.xaxis.label, ax.yaxis.label]
            + ax.get_xticklabels()
            + ax.get_yticklabels()
        ):
            item.set_fontsize(font_size_label)
            item.set_fontname(font_name)
    ax.title.set_fontsize(font_size_title)
    ax.title.set_fontname(font_name)

    if is_grid:
        ax.grid()

    if save_path is not None:
        save_path = save_path.replace("\\", "/")
        fig.savefig(save_path)
        plt.close()

    if is_show_fig:
        fig.show()

    if win_title:
        manager = plt.get_current_fig_manager()
        if manager is not None:
            manager.set_window_title(win_title)



# -*- coding: utf-8 -*-
from SciDataTool.Functions import AxisError
from numpy import tile, concatenate, negative, ones, append


def rebuild_symmetries(values, axis_index, symmetries):
    """Reconstructs the field of a Data object taking symmetries into account
    Parameters
    ----------
    values: ndarray
        ndarray of a field
    axis_index: int
        Index of the axis along which the symmetry is made
    symmetries: dict
        Dictionary of the symmetries along one axis
    Returns
    -------
    ndarray of the reconstructed field
    """
    if "period" in symmetries.keys():
        values = mytile(values, symmetries.get("period"), axis_index)
    elif "antiperiod" in symmetries.keys():
        values2 = concatenate((values, negative(values)), axis=axis_index)
        values3 = mytile(values2, symmetries.get("antiperiod") // 2, axis_index)
        if symmetries.get("antiperiod") % 2 == 1:
            values = concatenate((values3, values), axis=axis_index)
        else:
            values = values3
    return values


def rebuild_symmetries_axis(values, symmetries):
    """Reconstructs the field of a Data object taking symmetries into account
    Parameters
    ----------
    values: ndarray
        ndarray of a the axis values
    symmetries: dict
        Dictionary of the symmetries along the axis
    Returns
    -------
    ndarray of the reconstructed axis
    """
    values_new = values
    if "period" in symmetries.keys():
        for i in range(symmetries.get("period") - 1):
            if len(values) == 1:
                if "delta" in symmetries.keys():
                    values_new = append(
                        values_new, values_new[-1] + symmetries["delta"]
                    )
                else:
                    raise AxisError(
                        "ERROR: must provide delta for symmetries with one sample"
                    )
            else:
                values_new = concatenate(
                    (
                        values_new,
                        values + (values_new[-1] - values_new[-2]) + values_new[-1],
                    )
                )
    elif "antiperiod" in symmetries.keys():
        for i in range(symmetries.get("antiperiod") - 1):
            if len(values) == 1:
                if "delta" in symmetries.keys():
                    values_new = append(
                        values_new, values_new[-1] + symmetries["delta"]
                    )
                else:
                    raise AxisError(
                        "ERROR: must provide delta for symmetries with one sample"
                    )
            else:
                values_new = concatenate(
                    (
                        values_new,
                        values + (values_new[-1] - values_new[-2]) + values_new[-1],
                    )
                )
    return values_new


def mytile(values, n, axis_index):
    values_shape = values.shape
    reps = ones(len(values_shape), dtype=int)
    reps[axis_index] = n
    return tile(values, reps)



# -*- coding: utf-8 -*-
from SciDataTool.Functions import NormError, UnitError
from SciDataTool.Functions.conversions import convert, to_dB, to_dBA, to_noct
from numpy import apply_along_axis, abs as np_abs


def get_magnitude_along(
    self, *args, unit="SI", is_norm=False, axis_data=[], is_squeeze=True
):
    """Returns the ndarray of the magnitude of the FT, using conversions and symmetries if needed.
    Parameters
    ----------
    self: Data
        a Data object
    *args: list of strings
        List of axes requested by the user, their units and values (optional)
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    axis_data: list
        list of ndarray corresponding to user-input data
    Returns
    -------
    list of 1Darray of axes values, ndarray of magnitude values
    """
    if len(args) == 1 and type(args[0]) == tuple:
        args = args[0]  # if called from another script with *args
    return_dict = self.get_along(
        args, axis_data=axis_data, is_squeeze=is_squeeze, is_magnitude=True
    )
    values = return_dict[self.symbol]

    # 1/nth octave band
    for axis in return_dict["axes_list"]:
        if axis.name == "freqs" or axis.corr_name == "freqs":
            index = axis.index
            if axis.noct is not None:
                (values, foct) = apply_along_axis(
                    to_noct, index, values, return_dict["freqs"], noct=axis.noct
                )
                return_dict[axis.name] = foct
    # Convert into right unit (apart because of dBA conversion)
    if unit == self.unit or unit == "SI":
        if is_norm:
            try:
                values = values / self.normalizations.get("ref")
            except:
                raise NormError(
                    "ERROR: Reference value not specified for normalization"
                )
    elif unit == "dB":
        ref_value = 1.0
        if "ref" in self.normalizations.keys():
            ref_value *= self.normalizations.get("ref")
        values = to_dB(values, self.unit, ref_value)
    elif unit == "dBA":
        ref_value = 1.0
        if "ref" in self.normalizations.keys():
            ref_value *= self.normalizations.get("ref")
        if "freqs" in return_dict.keys():
            for axis in return_dict["axes_list"]:
                if axis.name == "freqs" or axis.corr_name == "freqs":
                    index = axis.index
            values = apply_along_axis(
                to_dBA, index, values, return_dict["freqs"], self.unit, ref_value
            )
        else:
            raise UnitError(
                "ERROR: dBA conversion only available for fft with frequencies"
            )
    elif unit in self.normalizations:
        values = values / self.normalizations.get(unit)
    else:
        values = convert(values, self.unit, unit)
    return_dict[self.symbol] = values
    return return_dict



# -*- coding: utf-8 -*-
from matplotlib.pyplot import subplots
import io
import pickle


def init_fig(fig=None, ax=None, shape="default", is_3d=False):
    """Get all the handle and legend of a figure or initialize them
    (for matplotlib)

    Parameters
    ----------
    fig : Matplotlib.figure.Figure
        The figure to get the handle from (can be None)
    ax : Matplotlib.axes.Axes object
        Axis on which to plot the data
    shape : str
        Shape of the figure: "default", "square" or "rectangle" for 20x10 figure
    is_3d : bool
        3D or 2D figure

    Returns
    -------
    (fig,axes,patch_leg,label_leg): Matplotlib.figure.Figure, matplotlib.axes._subplots.AxesSubplot, patch, string
        Figure handle, Axes Handle, List of legend patches, List of legend label

    """
    if fig is None:
        # Create a new figure with empty legend
        if shape == "rectangle":
            if is_3d:
                fig, axes = subplots(
                    tight_layout=True,
                    figsize=(8, 4),
                    subplot_kw=dict(projection="3d"),
                )
            else:
                fig, axes = subplots(tight_layout=True, figsize=(8, 4))
        elif shape == "square":
            if is_3d:
                fig, axes = subplots(
                    tight_layout=True,
                    figsize=(8, 8),
                    subplot_kw=dict(projection="3d"),
                )
            else:
                fig, axes = subplots(tight_layout=True, figsize=(8, 8))
        else:
            if is_3d:
                fig, axes = subplots(
                    tight_layout=True, subplot_kw=dict(projection="3d")
                )
            else:
                fig, axes = subplots(tight_layout=True)
        patch_leg, label_leg = [], []
    else:
        if ax is None:
            axes = fig.axes[0]
        else:
            axes = ax
        if axes.legend_ is None:
            # Empty legend
            patch_leg, label_leg = [], []
        else:
            # Get the symbol and label of all legend entry
            patch_leg = axes.legend_.get_patches()
            label_leg = [t.get_text() for t in axes.legend_.get_texts()]

    return (fig, axes, patch_leg, label_leg)


def init_subplot(fig=None, subplot_index=None, is_3d=False):
    """Initialize subplot (given position or automatic stacking)

    Parameters
    ----------
    fig : Matplotlib.figure.Figure
        The figure to get the handle from (can be None)
    subplot_index : int
        Index of the subplot, or None
    is_3d : bool
        3D or 2D figure

    Returns
    -------
    (fig,ax): Matplotlib.figure.Figure, matplotlib.axes._subplots.AxesSubplot
        Figure handle, Axes Handle

    """
    is_newfig = False
    if fig is None:
        is_newfig = True

    is_autostack = False
    if subplot_index is None:
        is_autostack = True

    (fig, axes, patch_leg, label_leg) = init_fig(fig, shape="rectangle", is_3d=is_3d)

    if not is_newfig and is_autostack:
        n = len(fig.axes)
        for i in range(n):
            fig.axes[i].change_geometry(n, 1, i)
        if is_3d:
            ax = fig.add_subplot(n, 1, n, projection="3d")
        else:
            ax = fig.add_subplot(n, 1, n)
    else:
        if subplot_index is None:
            subplot_index = 0
        ax = fig.axes[subplot_index]

    return fig, ax


def copy_fig(fig):
        """Copy the figure to avoid reference passing

        Returns
        -------
        Figure
            the new figure object
        """
        # Creating the buffer in which to pickle the fig
        buf = io.BytesIO()
        pickle.dump(fig, buf)
        buf.seek(0)
        # Loading the figure in a new figure
        fig_dock = pickle.load(buf)
        for i, ax in enumerate(fig_dock.axes):
            axes = ax
        return fig_dock


from cloudpickle import load


def load_pkl(file_path):
    """
    Load pyleecan object from pkl file

    Parameters
    ----------

    file_path: str
        file path

    Returns
    -------

    obj: Pyleecan object
    """
    with open(file_path, "rb") as file_:
        # file is a group
        obj = load(file_)

    return obj



# -*- coding: utf-8 -*-
from numpy import array_equal, ndarray


class FrozenClass(object):
    """A FrozenClass is designed to avoid adding or accessing wrong properties
    For instance w.wire instead of w.Wwire will throw a Frozen Error
    """

    __isfrozen = False  # Current state

    def __setattr__(self, key, value):
        """Overide to avoid the add of new properties outside of __init__
        Parameters
        ----------
        self : FrozenClass
            The class to set
        key : str
            The name of the property to set
        value : ?
            The value to set
        Raises
        ------
        FrozenError
            You can't set a new property if the class is frozen
        """
        # You can't add a new property to a frozen object
        if self.__isfrozen and not hasattr(self, key):
            raise FrozenError(
                self.__class__.__name__ + ' class has no "' + key + '" ' "property"
            )
        object.__setattr__(self, key, value)

    def _freeze(self):
        """After the call of this function, you can't add new properties to
        this FrozenClass
        Parameters
        ----------
        self : FrozenClass
            The FrozenClass object to freeze
        Returns
        -------
        None
        """
        self.__isfrozen = True

    def __eq__(self, other):
        """Two FrozenClass instance are equal if they have the same __dict__
        Parameters
        ----------
        self : FrozenClass
            A FrozenClass object
        other: FrozenClass
            Another FrozenClass object
        Returns
        -------
        is_equal: bool
            True if self == other
        """
        if isinstance(other, self.__class__):
            for attr in self.__dict__:
                if isinstance(self.__dict__[attr], ndarray):
                    if not array_equal(self.__dict__[attr], other.__dict__[attr]):
                        return False
                else:
                    if self.__dict__[attr] != other.__dict__[attr]:
                        return False
            return True
        else:
            return False

    def __ne__(self, other):
        """Call __eq__
        Parameters
        ----------
        self : FrozenClass
            A FrozenClass object
        other: FrozenClass
            Another FrozenClass object
        Returns
        -------
        is_not_equal: bool
            True if self != other
        """
        return not self.__eq__(other)


class FrozenError(Exception):
    """Throw when the class is frozen and a new property is set"""

    pass



# -*- coding: utf-8 -*-
from numpy import pi, sqrt, inf  # for eval
from SciDataTool.Functions import AxisError
from SciDataTool.Classes.RequestedAxis import RequestedAxis


def read_input_strings(args, axis_data):
    """Reads the string input into the "get_along" methods to define the axes
    Parameters
    ----------
    args: list
        list of string describing the requested axes
    axis_data: ndarray
        user-input values for the axes
    Returns
    -------
    list of axes data (axes_list)
    """
    axes_list = []
    for axis_str in args:
        unit = "SI"
        values = None
        indices = None
        input_data = None
        # Detect unit
        if "{" in axis_str:
            elems = axis_str.split("{")
            unit = elems[1].strip("}")
            axis_str = elems[0]
        # Detect normalization
        if "->" in axis_str:
            elems = axis_str.split("->")
            if "=" in elems[1]:
                unit = elems[1].split("=")[0]
            elif "[" in elems[1]:
                unit = elems[1].split("[")[0]
            elif ">" in elems[1]:
                unit = elems[1].split("[")[0]
            elif "<" in elems[1]:
                unit = elems[1].split("[")[0]
            else:
                unit = elems[1]
            name = elems[0]
            axis_str = axis_str.replace("->" + unit, "")
        # Detect rms sum
        if "rss" in axis_str:
            elems = axis_str.split("=rss")
            name = elems[0]
            extension = "rss"
        # Detect sum
        elif "sum" in axis_str:
            elems = axis_str.split("=sum")
            name = elems[0]
            extension = "sum"
        # Detect rms mean
        elif "rms" in axis_str:
            elems = axis_str.split("=rms")
            name = elems[0]
            extension = "rms"
        # Detect mean
        elif "mean" in axis_str:
            elems = axis_str.split("=mean")
            name = elems[0]
            extension = "mean"
        # Detect integrate
        elif "integrate" in axis_str:
            elems = axis_str.split("=integrate")
            name = elems[0]
            extension = "integrate"
        # Detect periods
        elif "oneperiod" in axis_str:
            elems = axis_str.split("[")
            name = elems[0]
            extension = "oneperiod"
        elif "antiperiod" in axis_str:
            elems = axis_str.split("[")
            name = elems[0]
            extension = "antiperiod"
        elif "smallestperiod" in axis_str:
            elems = axis_str.split("[")
            name = elems[0]
            extension = "smallestperiod"
        # Detect pattern
        elif "pattern" in axis_str:
            elems = axis_str.split("[")
            name = elems[0]
            extension = "pattern"
        # Detect axis_data input
        elif "axis_data" in axis_str:
            elems = axis_str.split("=axis_data")
            name = elems[0]
            extension = "axis_data"
            try:
                input_data = axis_data[name]
            except:
                raise AxisError("ERROR: No axis_data provided")
        # Detect above
        elif ">" in axis_str:
            elems = axis_str.split(">")
            init_str = elems[1]
            interval_init = eval(init_str)
            interval_final = inf
            name = elems[0]
            extension = "interval"
            input_data = [interval_init, interval_final]
        # Detect below
        elif "<" in axis_str:
            elems = axis_str.split("<")
            init_str = elems[1]
            interval_init = -inf
            interval_final = eval(init_str)
            name = elems[0]
            extension = "interval"
            input_data = [interval_init, interval_final]
        # Detect interval
        elif "=[" in axis_str:
            elems = axis_str.split("=[")
            elems2 = elems[1].split(",")
            if len(elems2) > 2:
                extension = "list"
                name = elems[0]
                input_data = [eval(elem.strip("]")) for elem in elems2]
            else:
                init_str = elems2[0]
                interval_init = eval(init_str)
                final_str = elems2[1].strip("]")
                interval_final = eval(final_str)
                name = elems[0]
                extension = "interval"
                input_data = [interval_init, interval_final]
        # Detect single value
        elif "=" in axis_str:
            elems = axis_str.split("=")
            name = elems[0]
            extension = "single"
            input_data = [eval(elems[1])]
        # Detect index input...
        elif "[" in axis_str:
            elems = axis_str.split("[")
            ind_str = elems[1].strip("]")
            name = elems[0]
            # Range of indices
            if ":" in ind_str:
                elems2 = ind_str.split(":")
                extension = "interval"
                indices = [i for i in range(int(elems2[0]), int(elems2[1]))]
            # List of indices
            elif "," in ind_str:
                extension = "list"
                indices = [int(x) for x in ind_str.split(",")]
            # Single index
            else:
                extension = "single"
                indices = [int(ind_str)]
        # Whole axis
        else:
            name = axis_str
            extension = "whole"
        # Detect 1/nth octave band
        if "oct" in unit:
            noct = int(unit.split("oct")[0].split("/")[1])
            unit = "SI"
        else:
            noct = None
        # RequestedAxis object creation
        axis = RequestedAxis(
            name=name,
            unit=unit,
            extension=extension,
            values=values,
            indices=indices,
            input_data=input_data,
            noct=noct,
        )
        axes_list.append(axis)
    return axes_list



# -*- coding: utf-8 -*-
from numpy import (
    array,
    linspace,
    argmin,
    take,
    isclose,
    isin,
    around,
    all,
    abs as np_abs,
    interp,
    where,
    zeros,
)
from scipy.interpolate import interp1d


def get_common_base(values1, values2, is_extrap=False, is_downsample=False):
    """Returns a common base for vectors values1 and values2
    Parameters
    ----------
    values1: list
        values of the first axis
    values2: list
        values of the second axis
    is_extrap: bool
        Boolean indicating if we want to keep the widest vector and extrapolate the other one
    is_downsample: bool
        Boolean indicating if we want to keep the smallest number of points and downsample the other one
    Returns
    -------
    list of the common axis values
    """
    if is_extrap:
        initial = min(values1[0], values2[0])
        final = max(values1[-1], values2[-1])
    else:
        initial = max(values1[0], values2[0])
        final = min(values1[-1], values2[-1])
    if is_downsample:
        number = min(
            len([i for i in values1 if i >= initial and i <= final]),
            len([i for i in values2 if i >= initial and i <= final]),
        )
    else:
        length1 = len([i for i in values1 if i >= initial and i <= final])
        length2 = len([i for i in values2 if i >= initial and i <= final])
        if length1 > length2:
            number = length1
            if initial not in values1:
                initial = values1[argmin(np_abs([i - initial for i in values1])) + 1]
            if final not in values1:
                final = values1[argmin(np_abs([i - final for i in values1])) - 1]
        else:
            number = length2
            if initial not in values2:
                initial = values2[argmin(np_abs([i - initial for i in values2])) + 1]
            if final not in values2:
                final = values2[argmin(np_abs([i - final for i in values2])) - 1]
    return linspace(initial, final, int(number), endpoint=True)


def get_interpolation(values, axis_values, new_axis_values, index):
    """Returns the interpolated field along one axis, given the new axis
    Parameters
    ----------
    values: ndarray
        array of a field
    axis_values: list
        values of the original axis
    new_axis_values: list
        values of the new axis
    index : int
        index of the axis
    Returns
    -------
    ndarray of the interpolated field
    """
    if str(axis_values) == "whole":  # Whole axis -> no interpolation
        return values
    elif len(new_axis_values) == 1:  # Single point -> use argmin
        idx = argmin(np_abs(axis_values - new_axis_values[0]))
        return take(values, [idx], axis=index)
    elif len(axis_values) == len(new_axis_values) and all(
        isclose(axis_values, new_axis_values, rtol=1e-03)
    ):  # Same axes -> no interpolation
        return values
    elif isin(
        new_axis_values, axis_values
    ).all():  # New axis is subset -> no interpolation
        indice_take = where(isin(axis_values, new_axis_values))[0]

        return take(
            values,
            indice_take,
            axis=index,
        )

    else:
        f = interp1d(axis_values, values, axis=index, fill_value="extrapolate")
        return f(new_axis_values)


def get_interpolation_step(values, axis_values, new_axis_values, index):
    """Returns the interpolated field along one axis, given the new axis
    Parameters
    ----------
    values: ndarray
        array of a field
    axis_values: list
        values of the original axis
    new_axis_values: list
        values of the new axis
    Returns
    -------
    ndarray of the interpolated field
    """
    if str(axis_values) == "whole":  # Whole axis -> no interpolation
        return values
    # elif len(new_axis_values) == 1:  # Single point -> use argmin
    #     idx = argmin(np_abs(axis_values - new_axis_values[0]))
    #     return take(values, [idx])
    elif len(axis_values) == len(new_axis_values) and all(
        isclose(axis_values, new_axis_values, rtol=1e-03)
    ):  # Same axes -> no interpolation
        return values
    elif isin(
        new_axis_values, axis_values
    ).all():  # New axis is subset -> no interpolation
        indice_take = where(isin(axis_values, new_axis_values))[0]

        return take(
            values,
            indice_take,
            axis=index,
        )

    else:
        new_shape = list(values.shape)
        new_shape[index] = len(new_axis_values)
        new_values = zeros(tuple(new_shape), dtype=values.dtype)
        if len(axis_values) == 1 or all(axis_values == axis_values[0]):
            for i in range(len(new_axis_values)):
                new_values[(slice(None),) * index + (i,)] = take(values, 0, axis=index)
            return new_values
        else:
            for i in range(len(new_axis_values)):
                for j in range(len(axis_values) - 1):
                    if isclose(
                        new_axis_values[i], axis_values[j], rtol=1e-03
                    ) and isclose(new_axis_values[i], axis_values[j + 1], rtol=1e-03):
                        new_values[(slice(None),) * index + (i,)] = (
                            take(values, j, axis=index)
                            + take(values, j + 1, axis=index)
                        ) / 2
                        break
                    elif (
                        new_axis_values[i] >= axis_values[j]
                        and new_axis_values[i] < axis_values[j + 1]
                        and not isclose(
                            new_axis_values[i], axis_values[j + 1], rtol=1e-03
                        )
                    ):
                        new_values[(slice(None),) * index + (i,)] = take(
                            values, j, axis=index
                        )
                        break
                    elif (
                        j == len(axis_values) - 2
                        and new_axis_values[i] == axis_values[j + 1]
                    ):
                        new_values[(slice(None),) * index + (i,)] = take(
                            values, j + 1, axis=index
                        )
                        break
            return array(new_values)



import numpy as np
import imageio
import matplotlib.pyplot as plt
from SciDataTool.Functions.Plot.init_fig import init_fig, copy_fig


def plot_2D_Data_Animated(
    self,
    animated_axis,
    *arg_list,
    nb_frames=50,
    fps=10,
    axis_data=None,
    radius=None,
    is_norm=False,
    unit="SI",
    component_list=None,
    data_list=[],
    legend_list=[],
    color_list=None,
    curve_colors=None,
    phase_colors=None,
    linestyles=None,
    linewidth_list=[2],
    save_path=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_disp_title=True,
    is_grid=True,
    is_auto_ticks=True,
    is_auto_range=False,
    fig=None,
    ax=None,
    barwidth=100,
    type_plot=None,
    fund_harm_dict=None,
    is_show_fig=None,
    win_title=None,
    thresh=0.02,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
    scale=None,
    width=0.005,
):
    """Plots a field as a function of time

    Parameters
    ----------
    self : Output
        an Output object
    Data_str : str
        name of the Data Object to plot (e.g. "mag.Br")
    animated_axis : str
        The field will be animated along this axis
    *arg_list : list of str
        arguments to specify which axes to plot
    nb_frames : int
        number of frames used to build the gif
    fps: int
        frames displayed per second
    radius : float
        radius used for quiver plot
    is_norm : bool
        boolean indicating if the field must be normalized
    unit : str
        unit in which to plot the field
    data_list : list
        list of Data objects to compare
    component_list : list
        list of component names to plot in separate figures
    legend_list : list
        list of legends to use for each Data object (including reference one) instead of data.name
    color_list : list
        list of colors to use for each Data object
    save_path : str
        full path of the png file where the figure is saved if save_path is not None
    x_min : float
        minimum value for the x-axis
    x_max : float
        maximum value for the x-axis
    y_min : float
        minimum value for the y-axis
    y_max : float
        maximum value for the y-axis
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_disp_title : bool
        boolean indicating if the title must be displayed
    is_grid : bool
        boolean indicating if the grid must be displayed
    is_auto_ticks : bool
        in fft, adjust ticks to freqs (deactivate if too close)
    is_auto_range : bool
        in fft, display up to 1% of max
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    barwidth : float
        barwidth scaling factor, only if type_plot = "bargraph"
    type_plot : str
        type of 2D graph : "curve", "bargraph", "barchart" or "quiver"
    fund_harm_dict : dict
        Dict containing axis name as key and frequency/order/wavenumber of fundamental harmonic as value to display fundamental harmonic in red in the fft
    is_show_fig : bool
        True to show figure after plot
    thresh : float
        threshold for automatic fft ticks
    scale : float
        arrow length factor in quiver. Be careful, if scale = None then there will be a normalization on the arrows on each frame
    width : float
        arrow width factor in quiver
    """

    # Special case of quiver plot
    if type_plot == "quiver":
        # list of png to make the gif
        images = list()

        # definition of the step along animated axis
        animated_values = self.get_axes(animated_axis)[0].get_values()
        value_max = np.max(animated_values)
        value_min = np.min(animated_values)
        variation_step = (value_max - value_min) / nb_frames

        # Getting the name of the gif
        save_path_gif = save_path.replace(".png", ".gif")
        # Copy_fig make a deep copy so that there is no overlaping
        deepcopy_fig = copy_fig(fig)

        # Params settings
        save_path = None
        is_show_fig = False

        # for value in animated_values:
        while value_min < value_max:
            # Call to plot_2D_Data of VectorField class, which manage the quiver case
            self.plot_2D_Data(
                *arg_list,
                animated_axis + "=" + str(value_min),
                axis_data=axis_data,
                radius=radius,
                is_norm=is_norm,
                unit=unit,
                component_list=component_list,
                data_list=data_list,
                legend_list=legend_list,
                color_list=color_list,
                curve_colors=curve_colors,
                phase_colors=phase_colors,
                linestyles=linestyles,
                linewidth_list=linewidth_list,
                save_path=save_path,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_disp_title=is_disp_title,
                is_grid=is_grid,
                is_auto_ticks=is_auto_ticks,
                is_auto_range=is_auto_range,
                fig=fig,
                ax=ax,
                barwidth=barwidth,
                type_plot=type_plot,
                fund_harm_dict=fund_harm_dict,
                is_show_fig=is_show_fig,
                win_title=win_title,
                thresh=thresh,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
                scale=scale,
                width=width,
            )
            fig.canvas.draw()
            image = np.frombuffer(fig.canvas.tostring_rgb(), dtype="uint8")
            image = image.reshape(fig.canvas.get_width_height()[::-1] + (3,))
            images.append(image)

            # Fig and ax reset to default for background
            fig = copy_fig(deepcopy_fig)
            ax = fig.axes[0]
            # While conidtion incrementation
            value_min += variation_step

        # Creating the gif
        plt.close(fig)
        imageio.mimsave(save_path_gif, images, format="GIF-PIL", fps=fps)

    else:

        # Call the plot on each component
        if component_list is None:  # default: extract all components
            component_list = self.components.keys()
        for i, comp in enumerate(component_list):
            # (fig, axes, patch_leg, label_leg) = init_fig(None, shape="rectangle")

            if save_path is not None and len(component_list) > 1:
                save_path_comp = (
                    save_path.split(".")[0] + "_" + comp + "." + save_path.split(".")[1]
                )
            else:
                save_path_comp = save_path

            self.components[comp].plot_2D_Data_Animated(
                animated_axis,
                *arg_list,
                nb_frames=nb_frames,
                fps=fps,
                axis_data=axis_data,
                is_norm=is_norm,
                unit=unit,
                data_list=[dat.components[comp] for dat in data_list],
                legend_list=legend_list,
                color_list=color_list,
                save_path=save_path_comp,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_disp_title=is_disp_title,
                is_grid=is_grid,
                is_auto_ticks=is_auto_ticks,
                is_auto_range=is_auto_range,
                fig=fig,
                ax=ax,
                barwidth=barwidth,
                type_plot=type_plot,
                fund_harm_dict=fund_harm_dict,
                is_show_fig=is_show_fig,
                thresh=thresh,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
            )



# -*- coding: utf-8 -*-
from SciDataTool.Functions.interpolations import get_common_base, get_interpolation
from numpy import squeeze, array


def compare_along(self, *args, data_list=[], unit="SI", is_norm=False):
    """Returns the ndarrays of both fields interpolated in the same axes, using conversions and symmetries if needed.
    Parameters
    ----------
    self: Data
        a Data object
    *args: list of strings
        List of axes requested by the user, their units and values (optional)
    data_list: list
        list of Data objects to compare
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    Returns
    -------
    list of 1Darray of axis values, ndarrays of fields
    """
    if data_list == []:
        return self.get_along(args, unit=unit, is_norm=is_norm)
    else:
        # Extract requested axes + field values
        results = self.get_along(args, unit=unit, is_norm=is_norm)
        values = results.pop(self.symbol)
        axes_list = results.pop("axes_list")
        axes_dict_other = results.pop("axes_dict_other")
        axes = results
        data_axis_values = []
        data_values = []
        return_dict = {}
        for data in data_list:
            results = data.get_along(args, unit=unit, is_norm=is_norm)
            data_values.append(results.pop(data.symbol))
            data_axis_values.append(results)
        # Get the common bases
        common_axis_values = {}
        for axis in axes.keys():
            common_axis_values[axis] = axes[axis]
            for i, data in enumerate(data_list):
                common_axis_values[axis] = get_common_base(
                    common_axis_values[axis], data_axis_values[i][axis]
                )
            # Interpolate over common axis values
            values = get_interpolation(values, axes[axis], common_axis_values[axis])
            for i, data in enumerate(data_list):
                data_values[i] = get_interpolation(
                    data_values[i],
                    data_axis_values[i][axis],
                    common_axis_values[axis],
                )
            return_dict[axis] = common_axis_values[axis]
        # Return axis and values
        return_dict[self.symbol] = values
        return_dict["axes_list"] = axes_list
        return_dict["axes_dict_other"] = axes_dict_other
        for i, data in enumerate(data_list):
            return_dict[data.symbol + "_" + str(i)] = data_values[i]
        return return_dict



# -*- coding: utf-8 -*-
from SciDataTool.Functions.interpolations import (
    get_interpolation,
    get_interpolation_step,
)
from numpy import apply_along_axis, squeeze


def interpolate(self, values, axes_list):
    """Returns the values of the field interpolated over the axes values.
    Parameters
    ----------
    self: Data
        a Data object
    values: ndarray
        array of the field
    axes_list: list
        a list of RequestedAxis objects
    Returns
    -------
    values: ndarray
        values of the field
    """

    # Interpolate over axis values
    for axis_requested in axes_list:
        if axis_requested.input_data is not None:
            if axis_requested.is_step:
                # values = apply_along_axis(
                #     get_interpolation_step,
                #     axis_requested.index,
                #     values,
                #     axis_requested.values,
                #     axis_requested.input_data,
                # )
                values = get_interpolation_step(
                    values,
                    axis_requested.values,
                    axis_requested.input_data,
                    axis_requested.index,
                )
            else:
                values = get_interpolation(
                    values,
                    axis_requested.values,
                    axis_requested.input_data,
                    axis_requested.index,
                )
            # Store new axis data into axis_requested.values
            axis_requested.values = axis_requested.input_data
    return values



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as file:
    requirements = file.readlines()
    install_requires = "".join(
        requirements
    ).splitlines()  # remove endline in each element

setuptools.setup(
    name="SciDataTool",
    version="1.4.2",
    author="Helene Toubin",
    author_email="helene.toubin@eomys.com",
    description="Scientific Data Tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Eomys/SciDataTool",
    download_url="https://github.com/Eomys/SciDataTool/archive/1.4.2.tar.gz",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=2.7",
    install_requires=install_requires,
)



# -*- coding: utf-8 -*-


def has_period(self):
    """Indicates if an axis has symmetries.
    Parameters
    ----------
    self: Data1D
        a Data1D object
    Returns
    -------
    Boolean
    """
    answer = False

    if "antiperiod" in self.symmetries:
        if self.symmetries["antiperiod"] > 1:
            answer = True
    elif "period" in self.symmetries:
        if self.symmetries["period"] > 1:
            answer = True

    return answer



# -*- coding: utf-8 -*-


def get_axes(self, *args):
    """Returns the list of axes.
    Parameters
    ----------
    self: Data
        a Data object
    args : list
        list of axes names
    Returns
    -------
    axes_list : list of axes (Data)
    """

    axes = self.components[list(self.components.keys())[0]].axes

    if len(args) > 0:
        axes_list = []
        for name in args:
            for axis in axes:
                if axis.name == name:
                    axes_list.append(axis)
    else:
        axes_list = axes

    return axes_list



# -*- coding: utf-8 -*-

from SciDataTool.Functions.Plot.plot_4D import plot_4D
from SciDataTool.Functions.Plot.plot_3D import plot_3D
from SciDataTool.Functions.Plot import unit_dict, norm_dict, axes_dict
from numpy import where, meshgrid, unique, max as np_max, min as np_min, array2string


def plot_3D_Data(
    self,
    *arg_list,
    axis_data=None,
    is_norm=False,
    unit="SI",
    save_path=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    z_min=None,
    z_max=None,
    z_range=None,
    is_auto_ticks=True,
    is_auto_range=True,
    is_2D_view=False,
    is_contour=False,
    is_same_size=False,
    N_stem=100,
    fig=None,
    ax=None,
    is_show_fig=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_logscale_z=False,
    thresh=None,
    is_switch_axes=False,
    colormap="RdBu_r",
    win_title=None,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
):
    """Plots a field as a function of two axes

    Parameters
    ----------
    data : Data
        a Data object
    *arg_list : list of str
        arguments to specify which axes to plot
    is_norm : bool
        boolean indicating if the field must be normalized
    unit : str
        unit in which to plot the field
    save_path : str
        full path including folder, name and extension of the file to save if save_path is not None
    x_min : float
        minimum value for the x-axis
    x_max : float
        maximum value for the x-axis
    y_min : float
        minimum value for the y-axis
    y_max : float
        maximum value for the y-axis
    z_min : float
        minimum value for the z-axis
    z_max : float
        maximum value for the z-axis
    z_range : float
        range to use for the z-axis
    is_auto_ticks : bool
        in fft, adjust ticks to freqs (deactivate if too close)
    is_auto_range : bool
        in fft, display up to 1% of max
    is_2D_view : bool
        True to plot Data in xy plane and put z as colormap
    is_contour : bool
        True to show contour line if is_fft = False and is_2D_view = True
    is_same_size : bool
        True to have all color blocks with same size in 2D view
    N_stem : int
        number of harmonics to plot (only for stem plots)
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    is_show_fig : bool
        True to show figure after plot
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_logscale_z : bool
        boolean indicating if the z-axis must be set in logarithmic scale
    thresh : float
        threshold for automatic fft ticks
    is_switch_axes : bool
        to switch x and y axes
    """

    if len(arg_list) == 1 and type(arg_list[0]) == tuple:
        arg_list = arg_list[0]  # if called from another script with *arg_list

    # Set unit
    if unit == "SI":
        unit = self.unit

    # Detect fft
    is_fft = False
    if any("wavenumber" in s for s in arg_list) or any("freqs" in s for s in arg_list):
        is_fft = True
        if "dB" in unit:
            if "ref" in self.normalizations:
                ref = self.normalizations["ref"]
            else:
                ref = 1
            unit_str = r"[" + unit + " re. " + str(ref) + "$" + self.unit + "$]"
        else:
            unit_str = r"$[" + unit + "]$"
        if self.symbol == "Magnitude":
            zlabel = "Magnitude " + unit_str
        else:
            zlabel = r"$|\widehat{" + self.symbol + "}|$ " + unit_str
        title1 = "FFT2 of " + self.name.lower() + " "
    else:
        if is_norm:
            zlabel = (
                r"$\frac{" + self.symbol + "}{" + self.symbol + "_0}\, [" + unit + "]$"
            )
        else:
            if self.symbol == "Magnitude":
                zlabel = "Magnitude " + r"$[" + unit + "]$"
            else:
                zlabel = r"$" + self.symbol + "\, [" + unit + "]$"
        title1 = "Surface plot of " + self.name.lower() + " "

    # Extract field and axes
    if is_fft:
        if is_2D_view:
            result = self.get_magnitude_along(
                arg_list, axis_data=axis_data, unit=unit, is_norm=is_norm
            )
        else:
            result = self.get_harmonics(
                N_stem,
                arg_list,
                axis_data=axis_data,
                unit=unit,
                is_norm=is_norm,
                is_flat=True,
            )
    else:
        result = self.get_along(arg_list, unit=unit, is_norm=is_norm)
    axes_list = result["axes_list"]
    axes_dict_other = result["axes_dict_other"]
    Xdata = result[axes_list[0].name]
    Ydata = result[axes_list[1].name]
    Zdata = result[self.symbol]
    if is_fft and not is_2D_view:
        X_flat = Xdata
        Y_flat = Ydata
        Z_flat = Zdata

    else:
        Y_map, X_map = meshgrid(Ydata, Xdata)
        X_flat = X_map.flatten()
        Y_flat = Y_map.flatten()
        Z_flat = Zdata.flatten()
    if x_min is None:
        x_min = np_min(Xdata)
    if x_max is None:
        x_max = np_max(Xdata)
    if y_min is None:
        y_min = np_min(Ydata)
    if y_max is None:
        y_max = np_max(Ydata)
    if z_range is None:
        if z_min is None:
            z_min = np_min(Zdata)
        if z_max is None:
            z_max = np_max(Zdata)
    else:
        if z_min is None and z_max is None:
            z_max = np_max(Zdata)
        if z_max is None:
            z_max = z_min + z_range
        if z_min is None:
            z_min = z_max - z_range

    # Build labels and titles
    axis = axes_list[0]
    if axis.name in axes_dict:
        name = axes_dict[axis.name]
    else:
        name = axis.name
    title2 = "over " + name.lower()
    if axis.unit == "SI":
        unit = unit_dict[axis.name]
        xlabel = name.capitalize() + " [" + unit + "]"
    elif axis.unit in norm_dict:
        xlabel = norm_dict[axis.unit]
    else:
        unit = axis.unit
        xlabel = name.capitalize() + " [" + unit + "]"
    if (
        axis.name == "angle"
        and axis.unit == "°"
        and round(np_max(axis.values) / 6) % 5 == 0
    ):
        xticks = [i * round(np_max(axis.values) / 6) for i in range(7)]
    else:
        xticks = None

    axis = axes_list[1]
    if axis.name in axes_dict:
        name = axes_dict[axis.name]
    else:
        name = axis.name
    title3 = " and " + axis.name.lower()
    if axis.unit == "SI":
        unit = unit_dict[axis.name]
        ylabel = name.capitalize() + " [" + unit + "]"
    elif axis.unit in norm_dict:
        ylabel = norm_dict[axis.unit]
    else:
        unit = axis.unit
        ylabel = name.capitalize() + " [" + unit + "]"
    if (
        axis.name == "angle"
        and axis.unit == "°"
        and round(np_max(axis.values) / 6) % 5 == 0
    ):
        yticks = [i * round(np_max(axis.values) / 6) for i in range(7)]
    else:
        yticks = None

    title4 = " for "
    for axis in axes_list[2:]:
        if axis.unit == "SI":
            unit = unit_dict[axis.name]
        elif axis.unit in norm_dict:
            unit = norm_dict[axis.unit]
        else:
            unit = axis.unit
        title4 += (
            axis.name
            + "="
            + array2string(
                result[axis.name],
                formatter={"float_kind": "{:.3g}".format},
            )
            .replace(" ", ", ")
            .replace("[", "")
            .replace("]", "")
            + " ["
            + unit
            + "], "
        )
    title5 = ""
    for axis_name in axes_dict_other:
        title5 += (
            axis_name
            + "="
            + array2string(
                axes_dict_other[axis_name][0],
                formatter={"float_kind": "{:.3g}".format},
            ).replace(" ", ", ")
            + " ["
            + axes_dict_other[axis_name][1]
            + "], "
        )

    if title4 == " for " and title5 == "":
        title4 = ""

    title = title1 + title2 + title3 + title4 + title5
    title = title.rstrip(", ")

    if is_fft:

        if thresh is None:
            if self.normalizations is not None and "ref" in self.normalizations.keys():
                thresh = self.normalizations["ref"]
            else:
                thresh = 0.02

        indices = where(Z_flat > abs(thresh * np_max(Zdata)))[0]
        xticks = unique(X_flat[indices])
        yticks = unique(Y_flat[indices])
        if is_auto_range:
            if len(xticks) > 0:
                x_min = -0.1 * xticks[-1]
                x_max = xticks[-1] * 1.1
            if len(yticks) > 0:
                y_min = yticks[0] * 1.1
                y_max = yticks[-1] * 1.1
        else:
            x_min = x_min - x_max * 0.05
            x_max = x_max * 1.05
            y_min = y_min - y_max * 0.2
            y_max = y_max * 1.2
        if not is_auto_ticks:
            xticks = None
            yticks = None
        if is_2D_view:
            plot_4D(
                X_flat,
                Y_flat,
                Z_flat,
                Sdata=None,
                is_same_size=is_same_size,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                z_max=z_max,
                z_min=z_min,
                title=title,
                xticks=xticks,
                yticks=yticks,
                xlabel=xlabel,
                ylabel=ylabel,
                zlabel=zlabel,
                fig=fig,
                ax=ax,
                type_plot="scatter",
                save_path=save_path,
                is_show_fig=is_show_fig,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_logscale_z=is_logscale_z,
                is_switch_axes=is_switch_axes,
                colormap=colormap,
                win_title=win_title,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
                is_grid=True,
            )
        else:
            plot_3D(
                X_flat,
                Y_flat,
                Z_flat,
                fig=fig,
                ax=ax,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                z_min=0,
                z_max=z_max,
                title=title,
                xticks=xticks,
                yticks=yticks,
                xlabel=xlabel,
                ylabel=ylabel,
                zlabel=zlabel,
                type_plot="stem",
                save_path=save_path,
                is_show_fig=is_show_fig,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_logscale_z=is_logscale_z,
                is_switch_axes=is_switch_axes,
                colormap=colormap,
                win_title=win_title,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
            )
    else:
        if is_2D_view:
            plot_3D(
                Xdata,
                Ydata,
                Zdata,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                z_max=z_max,
                z_min=z_min,
                xlabel=xlabel,
                ylabel=ylabel,
                zlabel=zlabel,
                title=title,
                xticks=xticks,
                yticks=yticks,
                fig=fig,
                ax=ax,
                type_plot="pcolor",
                is_contour=is_contour,
                save_path=save_path,
                is_show_fig=is_show_fig,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_logscale_z=is_logscale_z,
                is_switch_axes=is_switch_axes,
                colormap=colormap,
                win_title=win_title,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
            )
        else:
            plot_3D(
                X_map,
                Y_map,
                Zdata,
                fig=fig,
                ax=ax,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                z_min=z_min,
                z_max=z_max,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                zlabel=zlabel,
                yticks=yticks,
                type_plot="surf",
                save_path=save_path,
                is_show_fig=is_show_fig,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_logscale_z=is_logscale_z,
                is_switch_axes=is_switch_axes,
                colormap=colormap,
                win_title=win_title,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
            )



# -*- coding: utf-8 -*-
from SciDataTool.Functions.conversions import convert
from SciDataTool.Functions.symmetries import rebuild_symmetries_axis
from SciDataTool.Functions import AxisError


def get_values(
    self, unit="SI", is_oneperiod=False, is_antiperiod=False, is_smallestperiod=False
):
    """Returns the vector 'axis' taking symmetries into account.
    Parameters
    ----------
    self: Data1D
        a Data1D object
    unit: str
        requested unit
    is_oneperiod: bool
        return values on a single period
    is_antiperiod: bool
        return values on a semi period (only for antiperiodic signals)
    Returns
    -------
    Vector of axis values
    """
    values = self.values

    # Unit conversion
    if unit != "SI" and unit != self.unit:
        values = convert(values, self.unit, unit)

    # Ignore symmetries if fft axis
    if self.name == "freqs" or self.name == "wavenumber":
        is_smallestperiod = True

    # Rebuild symmetries
    if is_smallestperiod:
        return values
    elif is_antiperiod:
        if "antiperiod" in self.symmetries:
            return values
        else:
            raise AxisError("ERROR: axis has no antiperiodicity")
    elif is_oneperiod:
        if "antiperiod" in self.symmetries:
            nper = self.symmetries["antiperiod"]
            self.symmetries["antiperiod"] = 2
            values = rebuild_symmetries_axis(values, self.symmetries)
            self.symmetries["antiperiod"] = nper
            return values
        elif "period" in self.symmetries:
            return values
        else:
            return values
    else:
        values = rebuild_symmetries_axis(values, self.symmetries)
        return values



from h5py import File, Group
from numpy import bool_, int64, string_, array
from cloudpickle import loads


def construct_dict_from_group(group):
    """
    construct_dict_from_group create a dictionnary and extract datasets and groups from the group

    Parameters
    ----------
    group: h5py.Group
        group to browse

    Returns
    -------
    dict_ : dict
        created dict containing the group data
    """
    dict_ = {}

    # List split to load
    if "length_list" in group.attrs.keys():
        list_ = []

        for i in range(group.attrs["length_list"]):
            if hasattr(group["list_" + str(i)], "items"):  # Group in list
                list_.append(construct_dict_from_group(group["list_" + str(i)]))
            else:  # Dataset
                dataset = group["list_" + str(i)]
                value = dataset[()]
                if "array_list" in dataset.attrs.keys():  # List saved as an array
                    value = value.tolist()
                elif value == "NoneValue":  # Handle None values
                    value = None
                elif isinstance(value, bool_):  # bool
                    value = bool(value)
                elif isinstance(value, int64):  # float
                    value = float(value)
                elif isinstance(value, string_):  # String
                    value = value.decode("ISO-8859-2")

                list_.append(value)
        return list_
    else:
        for key, val in group.items():
            # Check if val is a group or a dataset
            if isinstance(val, Group):  # Group
                # Call the function recursively to load group
                dict_[key] = construct_dict_from_group(val)
            else:  # Dataset
                value = val[()]
                if "array_list" in val.attrs.keys():  # List saved as an array
                    value = value.tolist()
                elif value == "NoneValue":  # Handle None values
                    value = None
                elif isinstance(value, bool_):  # bool
                    value = bool(value)
                elif isinstance(value, int64):  # float
                    value = float(value)
                elif isinstance(value, string_):  # String
                    value = value.decode("ISO-8859-2")
                dict_[key] = value
        return dict_


def load_hdf5(file_path):
    """
    Load pyleecan object from h5 file

    Parameters
    ----------

    file_path: str
        file path

    Returns
    -------

    file_path: str
    obj_dict: dict
        dictionnary to instanciate Pyleecan obj
    """
    with File(file_path, "r") as file:
        # file is a group
        obj_dict = construct_dict_from_group(file)

    return file_path, obj_dict



# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/DataND.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/SciDataTool/tree/master/SciDataTool/Methods//DataND
"""

from os import linesep
from sys import getsizeof
from ._check import set_array, check_var, raise_
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from .Data import Data

# Import all class method
# Try/catch to remove unnecessary dependencies in unused method
try:
    from ..Methods.DataND._set_values import _set_values
except ImportError as error:
    _set_values = error

try:
    from ..Methods.DataND.comp_axes import comp_axes
except ImportError as error:
    comp_axes = error

try:
    from ..Methods.DataND.compare_along import compare_along
except ImportError as error:
    compare_along = error

try:
    from ..Methods.DataND.compare_magnitude_along import compare_magnitude_along
except ImportError as error:
    compare_magnitude_along = error

try:
    from ..Methods.DataND.compare_phase_along import compare_phase_along
except ImportError as error:
    compare_phase_along = error

try:
    from ..Methods.DataND.compress import compress
except ImportError as error:
    compress = error

try:
    from ..Methods.DataND.convert import convert
except ImportError as error:
    convert = error

try:
    from ..Methods.DataND.export_along import export_along
except ImportError as error:
    export_along = error

try:
    from ..Methods.DataND.extract_slices import extract_slices
except ImportError as error:
    extract_slices = error

try:
    from ..Methods.DataND.extract_slices_fft import extract_slices_fft
except ImportError as error:
    extract_slices_fft = error

try:
    from ..Methods.DataND.get_along import get_along
except ImportError as error:
    get_along = error

try:
    from ..Methods.DataND.get_axes import get_axes
except ImportError as error:
    get_axes = error

try:
    from ..Methods.DataND.get_data_along import get_data_along
except ImportError as error:
    get_data_along = error

try:
    from ..Methods.DataND.get_field import get_field
except ImportError as error:
    get_field = error

try:
    from ..Methods.DataND.get_harmonics import get_harmonics
except ImportError as error:
    get_harmonics = error

try:
    from ..Methods.DataND.get_magnitude_along import get_magnitude_along
except ImportError as error:
    get_magnitude_along = error

try:
    from ..Methods.DataND.get_phase_along import get_phase_along
except ImportError as error:
    get_phase_along = error

try:
    from ..Methods.DataND.has_period import has_period
except ImportError as error:
    has_period = error

try:
    from ..Methods.DataND.interpolate import interpolate
except ImportError as error:
    interpolate = error

try:
    from ..Methods.DataND.plot_2D_Data import plot_2D_Data
except ImportError as error:
    plot_2D_Data = error

try:
    from ..Methods.DataND.plot_2D_Data_Animated import plot_2D_Data_Animated
except ImportError as error:
    plot_2D_Data_Animated = error

try:
    from ..Methods.DataND.plot_3D_Data import plot_3D_Data
except ImportError as error:
    plot_3D_Data = error

try:
    from ..Methods.DataND.rebuild_symmetries import rebuild_symmetries
except ImportError as error:
    rebuild_symmetries = error

try:
    from ..Methods.DataND.set_Ftparameters import set_Ftparameters
except ImportError as error:
    set_Ftparameters = error


from numpy import array, array_equal
from ._check import InitUnKnowClassError


class DataND(Data):
    """Abstract class for fields (time or frequency domain)"""

    VERSION = 1

    # Check ImportError to remove unnecessary dependencies in unused method
    # cf Methods.DataND._set_values
    if isinstance(_set_values, ImportError):
        _set_values = property(
            fget=lambda x: raise_(
                ImportError("Can't use DataND method _set_values: " + str(_set_values))
            )
        )
    else:
        _set_values = _set_values
    # cf Methods.DataND.comp_axes
    if isinstance(comp_axes, ImportError):
        comp_axes = property(
            fget=lambda x: raise_(
                ImportError("Can't use DataND method comp_axes: " + str(comp_axes))
            )
        )
    else:
        comp_axes = comp_axes
    # cf Methods.DataND.compare_along
    if isinstance(compare_along, ImportError):
        compare_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method compare_along: " + str(compare_along)
                )
            )
        )
    else:
        compare_along = compare_along
    # cf Methods.DataND.compare_magnitude_along
    if isinstance(compare_magnitude_along, ImportError):
        compare_magnitude_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method compare_magnitude_along: "
                    + str(compare_magnitude_along)
                )
            )
        )
    else:
        compare_magnitude_along = compare_magnitude_along
    # cf Methods.DataND.compare_phase_along
    if isinstance(compare_phase_along, ImportError):
        compare_phase_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method compare_phase_along: "
                    + str(compare_phase_along)
                )
            )
        )
    else:
        compare_phase_along = compare_phase_along
    # cf Methods.DataND.compress
    if isinstance(compress, ImportError):
        compress = property(
            fget=lambda x: raise_(
                ImportError("Can't use DataND method compress: " + str(compress))
            )
        )
    else:
        compress = compress
    # cf Methods.DataND.convert
    if isinstance(convert, ImportError):
        convert = property(
            fget=lambda x: raise_(
                ImportError("Can't use DataND method convert: " + str(convert))
            )
        )
    else:
        convert = convert
    # cf Methods.DataND.export_along
    if isinstance(export_along, ImportError):
        export_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method export_along: " + str(export_along)
                )
            )
        )
    else:
        export_along = export_along
    # cf Methods.DataND.extract_slices
    if isinstance(extract_slices, ImportError):
        extract_slices = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method extract_slices: " + str(extract_slices)
                )
            )
        )
    else:
        extract_slices = extract_slices
    # cf Methods.DataND.extract_slices_fft
    if isinstance(extract_slices_fft, ImportError):
        extract_slices_fft = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method extract_slices_fft: "
                    + str(extract_slices_fft)
                )
            )
        )
    else:
        extract_slices_fft = extract_slices_fft
    # cf Methods.DataND.get_along
    if isinstance(get_along, ImportError):
        get_along = property(
            fget=lambda x: raise_(
                ImportError("Can't use DataND method get_along: " + str(get_along))
            )
        )
    else:
        get_along = get_along
    # cf Methods.DataND.get_axes
    if isinstance(get_axes, ImportError):
        get_axes = property(
            fget=lambda x: raise_(
                ImportError("Can't use DataND method get_axes: " + str(get_axes))
            )
        )
    else:
        get_axes = get_axes
    # cf Methods.DataND.get_data_along
    if isinstance(get_data_along, ImportError):
        get_data_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method get_data_along: " + str(get_data_along)
                )
            )
        )
    else:
        get_data_along = get_data_along
    # cf Methods.DataND.get_field
    if isinstance(get_field, ImportError):
        get_field = property(
            fget=lambda x: raise_(
                ImportError("Can't use DataND method get_field: " + str(get_field))
            )
        )
    else:
        get_field = get_field
    # cf Methods.DataND.get_harmonics
    if isinstance(get_harmonics, ImportError):
        get_harmonics = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method get_harmonics: " + str(get_harmonics)
                )
            )
        )
    else:
        get_harmonics = get_harmonics
    # cf Methods.DataND.get_magnitude_along
    if isinstance(get_magnitude_along, ImportError):
        get_magnitude_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method get_magnitude_along: "
                    + str(get_magnitude_along)
                )
            )
        )
    else:
        get_magnitude_along = get_magnitude_along
    # cf Methods.DataND.get_phase_along
    if isinstance(get_phase_along, ImportError):
        get_phase_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method get_phase_along: " + str(get_phase_along)
                )
            )
        )
    else:
        get_phase_along = get_phase_along
    # cf Methods.DataND.has_period
    if isinstance(has_period, ImportError):
        has_period = property(
            fget=lambda x: raise_(
                ImportError("Can't use DataND method has_period: " + str(has_period))
            )
        )
    else:
        has_period = has_period
    # cf Methods.DataND.interpolate
    if isinstance(interpolate, ImportError):
        interpolate = property(
            fget=lambda x: raise_(
                ImportError("Can't use DataND method interpolate: " + str(interpolate))
            )
        )
    else:
        interpolate = interpolate
    # cf Methods.DataND.plot_2D_Data
    if isinstance(plot_2D_Data, ImportError):
        plot_2D_Data = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method plot_2D_Data: " + str(plot_2D_Data)
                )
            )
        )
    else:
        plot_2D_Data = plot_2D_Data
    # cf Methods.DataND.plot_2D_Data_Animated
    if isinstance(plot_2D_Data_Animated, ImportError):
        plot_2D_Data_Animated = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method plot_2D_Data_Animated: "
                    + str(plot_2D_Data_Animated)
                )
            )
        )
    else:
        plot_2D_Data_Animated = plot_2D_Data_Animated
    # cf Methods.DataND.plot_3D_Data
    if isinstance(plot_3D_Data, ImportError):
        plot_3D_Data = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method plot_3D_Data: " + str(plot_3D_Data)
                )
            )
        )
    else:
        plot_3D_Data = plot_3D_Data
    # cf Methods.DataND.rebuild_symmetries
    if isinstance(rebuild_symmetries, ImportError):
        rebuild_symmetries = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method rebuild_symmetries: "
                    + str(rebuild_symmetries)
                )
            )
        )
    else:
        rebuild_symmetries = rebuild_symmetries
    # cf Methods.DataND.set_Ftparameters
    if isinstance(set_Ftparameters, ImportError):
        set_Ftparameters = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataND method set_Ftparameters: " + str(set_Ftparameters)
                )
            )
        )
    else:
        set_Ftparameters = set_Ftparameters
    # save and copy methods are available in all object
    save = save
    copy = copy

    def __init__(
        self,
        axes=None,
        FTparameters=-1,
        values=None,
        is_real=True,
        symbol="",
        name="",
        unit="",
        normalizations=-1,
        init_dict=None,
        init_str=None,
    ):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for SciDataTool type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionnary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for SciDataTool Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "axes" in list(init_dict.keys()):
                axes = init_dict["axes"]
            if "FTparameters" in list(init_dict.keys()):
                FTparameters = init_dict["FTparameters"]
            if "values" in list(init_dict.keys()):
                values = init_dict["values"]
            if "is_real" in list(init_dict.keys()):
                is_real = init_dict["is_real"]
            if "symbol" in list(init_dict.keys()):
                symbol = init_dict["symbol"]
            if "name" in list(init_dict.keys()):
                name = init_dict["name"]
            if "unit" in list(init_dict.keys()):
                unit = init_dict["unit"]
            if "normalizations" in list(init_dict.keys()):
                normalizations = init_dict["normalizations"]
        # Set the properties (value check and convertion are done in setter)
        self.axes = axes
        self.FTparameters = FTparameters
        self.values = values
        self.is_real = is_real
        # Call Data init
        super(DataND, self).__init__(
            symbol=symbol, name=name, unit=unit, normalizations=normalizations
        )
        # The class is frozen (in Data init), for now it's impossible to
        # add new properties

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        DataND_str = ""
        # Get the properties inherited from Data
        DataND_str += super(DataND, self).__str__()
        DataND_str += "axes = " + str(self.axes) + linesep + linesep
        DataND_str += "FTparameters = " + str(self.FTparameters) + linesep
        DataND_str += (
            "values = "
            + linesep
            + str(self.values).replace(linesep, linesep + "\t")
            + linesep
            + linesep
        )
        DataND_str += "is_real = " + str(self.is_real) + linesep
        return DataND_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False

        # Check the properties inherited from Data
        if not super(DataND, self).__eq__(other):
            return False
        if other.axes != self.axes:
            return False
        if other.FTparameters != self.FTparameters:
            return False
        if not array_equal(other.values, self.values):
            return False
        if other.is_real != self.is_real:
            return False
        return True

    def compare(self, other, name="self"):
        """Compare two objects and return list of differences"""

        if type(other) != type(self):
            return ["type(" + name + ")"]
        diff_list = list()

        # Check the properties inherited from Data
        diff_list.extend(super(DataND, self).compare(other, name=name))
        if (other.axes is None and self.axes is not None) or (
            other.axes is not None and self.axes is None
        ):
            diff_list.append(name + ".axes None mismatch")
        elif self.axes is None:
            pass
        elif len(other.axes) != len(self.axes):
            diff_list.append("len(" + name + ".axes)")
        else:
            for ii in range(len(other.axes)):
                diff_list.extend(
                    self.axes[ii].compare(
                        other.axes[ii], name=name + ".axes[" + str(ii) + "]"
                    )
                )
        if other._FTparameters != self._FTparameters:
            diff_list.append(name + ".FTparameters")
        if not array_equal(other.values, self.values):
            diff_list.append(name + ".values")
        if other._is_real != self._is_real:
            diff_list.append(name + ".is_real")
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object

        # Get size of the properties inherited from Data
        S += super(DataND, self).__sizeof__()
        if self.axes is not None:
            for value in self.axes:
                S += getsizeof(value)
        if self.FTparameters is not None:
            for key, value in self.FTparameters.items():
                S += getsizeof(value) + getsizeof(key)
        S += getsizeof(self.values)
        S += getsizeof(self.is_real)
        return S

    def as_dict(self, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        Optional keyword input parameter is for internal use only
        and may prevent json serializability.
        """

        # Get the properties inherited from Data
        DataND_dict = super(DataND, self).as_dict(**kwargs)
        if self.axes is None:
            DataND_dict["axes"] = None
        else:
            DataND_dict["axes"] = list()
            for obj in self.axes:
                if obj is not None:
                    DataND_dict["axes"].append(obj.as_dict())
                else:
                    DataND_dict["axes"].append(None)
        DataND_dict["FTparameters"] = (
            self.FTparameters.copy() if self.FTparameters is not None else None
        )
        if self.values is None:
            DataND_dict["values"] = None
        else:
            DataND_dict["values"] = self.values.tolist()
        DataND_dict["is_real"] = self.is_real
        # The class name is added to the dict for deserialisation purpose
        # Overwrite the mother class name
        DataND_dict["__class__"] = "DataND"
        return DataND_dict

    def _set_None(self):
        """Set all the properties to None (except SciDataTool object)"""

        self.axes = None
        self.FTparameters = None
        self.values = None
        self.is_real = None
        # Set to None the properties inherited from Data
        super(DataND, self)._set_None()

    def _get_axes(self):
        """getter of axes"""
        if self._axes is not None:
            for obj in self._axes:
                if obj is not None:
                    obj.parent = self
        return self._axes

    def _set_axes(self, value):
        """setter of axes"""
        if type(value) is list:
            for ii, obj in enumerate(value):
                if type(obj) is dict:
                    class_obj = import_class(
                        "SciDataTool.Classes", obj.get("__class__"), "axes"
                    )
                    value[ii] = class_obj(init_dict=obj)
                if value[ii] is not None:
                    value[ii].parent = self
        if value == -1:
            value = list()
        check_var("axes", value, "[Data]")
        self._axes = value

    axes = property(
        fget=_get_axes,
        fset=_set_axes,
        doc=u"""List of the Data1D objects corresponding to the axes

        :Type: [SciDataTool.Classes.Data]
        """,
    )

    def _get_FTparameters(self):
        """getter of FTparameters"""
        return self._FTparameters

    def _set_FTparameters(self, value):
        """setter of FTparameters"""
        if type(value) is int and value == -1:
            value = dict()
        check_var("FTparameters", value, "dict")
        self._FTparameters = value

    FTparameters = property(
        fget=_get_FTparameters,
        fset=_set_FTparameters,
        doc=u"""Tunable parameters for the Fourier Transforms

        :Type: dict
        """,
    )

    def _get_values(self):
        """getter of values"""
        return self._values

    values = property(
        fget=_get_values,
        fset=_set_values,
        doc=u"""Values of the field

        :Type: ndarray
        """,
    )

    def _get_is_real(self):
        """getter of is_real"""
        return self._is_real

    def _set_is_real(self, value):
        """setter of is_real"""
        check_var("is_real", value, "bool")
        self._is_real = value

    is_real = property(
        fget=_get_is_real,
        fset=_set_is_real,
        doc=u"""To indicate if the signal is real (use only positive frequencies)

        :Type: bool
        """,
    )



# -*- coding: utf-8 -*-

from itertools import repeat

import matplotlib.pyplot as plt

from numpy import (
    ceil,
    argmin,
    abs,
    squeeze,
    split,
    ndarray,
    cumsum,
    zeros,
    shape,
    max as np_max,
)

from SciDataTool.Functions.Plot.init_fig import init_fig
from SciDataTool.Functions.Plot import COLORS, LINESTYLES


def plot_2D(
    Xdatas,
    Ydatas,
    legend_list=[""],
    color_list=None,
    linestyle_list=None,
    linewidth_list=[2],
    title="",
    xlabel="",
    ylabel="",
    fig=None,
    ax=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_disp_title=True,
    is_grid=True,
    type_plot="curve",
    fund_harm=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    xticks=None,
    save_path=None,
    barwidth=5,
    is_show_fig=None,
    win_title=None,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
    is_show_legend=True,
    scale_units="x",
    scale=None,
    width=0.005,
    symbol=".",
):
    """Plots a 2D graph (curve, bargraph or barchart) comparing fields in Ydatas

    Parameters
    ----------
    Xdatas : ndarray
        array of x-axis values
    Ydatas : list
        list of y-axes values
    legend_list : list
        list of legends
    color_list : list
        list of colors to use for each curve
    linewidth_list : list
        list of line width to use for each curve
    title : str
        title of the graph
    xlabel : str
        label for the x-axis
    ylabel : str
        label for the y-axis
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_disp_title : bool
        boolean indicating if the title must be displayed
    is_grid : bool
        boolean indicating if the grid must be displayed
    type_plot : str
        type of 2D graph : "curve", "bargraph", "barchart" or "quiver"
    fund_harm : float
        frequency/order/wavenumber of the fundamental harmonic that must be displayed in red in the fft
    x_min : float
        minimum value for the x-axis
    x_max : float
        maximum value for the x-axis
    y_min : float
        minimum value for the y-axis
    y_max : float
        maximum value for the y-axis
    xticks : list
        list of ticks to use for the x-axis
    save_path : str
        full path including folder, name and extension of the file to save if save_path is not None
    barwidth : float
        barwidth scaling factor, only if type_plot = "bargraph"
    is_show_fig : bool
        True to show figure after plot
    win_title : str
        Title of the plot window
    scale_units : str
        arrow lenght scale factor reference {'width', 'height', 'dots', 'pouces', 'x', 'y', 'xy'}
    scale : float
        arrow lenght factor
    width : float
        arrow width factor
    """

    # Set is_show_fig if is None
    if is_show_fig is None:
        is_show_fig = True if fig is None else False

    # Set figure if needed
    if fig is None and ax is None:
        (fig, ax, _, _) = init_fig(fig=None, shape="rectangle")

    # Number of curves on a axe
    ndatas = len(Ydatas)

    # Retrocompatibility
    if isinstance(Xdatas, ndarray):
        Xdatas = [Xdatas]

    if len(Xdatas) == 1:
        i_Xdatas = [0 for i in range(ndatas)]
    else:
        i_Xdatas = range(ndatas)

    # Expand default arguments
    if color_list is None:
        color_list = COLORS
    if linestyle_list is None:
        linestyle_list = ["solid" for i in range(ndatas)]
    if len(color_list) < ndatas:
        # Repeat colors and change linestyles
        l = len(color_list)
        n = int(ceil(ndatas / l))
        color_list = color_list * n
        if len(linestyle_list) < n:
            # Expand linestyles
            linestyle_list = linestyle_list + [
                linestyle for linestyle in LINESTYLES if linestyle not in linestyle_list
            ]
        if len(linestyle_list) < n:
            # Linestyles is still too small -> repeat
            m = int(ceil(n / len(linestyle_list)))
            linestyle_list = linestyle_list * m
        linestyles = []
        for i in range(n):
            linestyles += [linestyle_list[i] for j in range(l)]
        linestyle_list = linestyles
    if len(linestyle_list) < ndatas:
        # Repeat linestyles
        m = int(ceil(ndatas / len(linestyle_list)))
        linestyle_list = linestyle_list * m
    if len(linewidth_list) < ndatas:
        # Repeat linewidths
        m = int(ceil(ndatas / len(linewidth_list)))
        linewidth_list = linewidth_list * m
    if (len(legend_list) == 1 and legend_list[0] == "") or len(legend_list) == 0:
        no_legend = True
        legend_list = ["" for i in range(ndatas)]
    else:
        no_legend = False

    # Plot
    if type_plot == "curve":
        for i in range(ndatas):
            ax.plot(
                Xdatas[i_Xdatas[i]],
                Ydatas[i],
                color=color_list[i],
                label=legend_list[i],
                linewidth=linewidth_list[i],
                ls=linestyle_list[i],
                picker=True,
                pickradius=5,
            )
        if xticks is not None:
            ax.xaxis.set_ticks(xticks)
    elif type_plot == "bargraph":
        positions = range(-ndatas + 1, ndatas, 2)
        if x_max is not None:
            width = x_max / barwidth
        else:
            width = Xdatas[i_Xdatas[0]][-1] / barwidth
        for i in range(ndatas):
            # width = (Xdatas[i_Xdatas[i]][1] - Xdatas[i_Xdatas[i]][0]) / ndatas
            barlist = ax.bar(
                Xdatas[i_Xdatas[i]] + positions[i] * width / (2 * ndatas),
                Ydatas[i],
                color=color_list[i],
                width=width,
                label=legend_list[i],
                picker=True,
            )
            if fund_harm is not None:  # Find fundamental
                imax = argmin(abs(Xdatas[i] - fund_harm))
                if ndatas < 3 and len(color_list) > 2:
                    fund_color = color_list[2]
                else:
                    fund_color = "k"
                barlist[imax].set_edgecolor(fund_color)
                barlist[imax].set_facecolor(fund_color)

        if xticks is not None:
            ax.xaxis.set_ticks(xticks)
            plt.xticks(rotation=90, ha="center", va="top")

    elif type_plot == "barchart":
        for i in range(ndatas):
            if i == 0:
                ax.bar(
                    range(len(Xdatas[i_Xdatas[i]])),
                    Ydatas[i],
                    color=color_list[i],
                    width=0.5,
                    label=legend_list[i],
                    picker=True,
                )
            else:
                ax.bar(
                    range(len(Xdatas[i_Xdatas[i]])),
                    Ydatas[i],
                    edgecolor=color_list[i],
                    width=0.5,
                    fc="None",
                    lw=1,
                    label=legend_list[i],
                    picker=True,
                )
        plt.xticks(
            range(len(Xdatas[i_Xdatas[i]])),
            ["{:.2f}".format(f) for f in Xdatas[i_Xdatas[i]]],
            rotation=90,
        )

    elif type_plot == "quiver":
        for i in range(ndatas):
            x = [e[0] for e in Xdatas[i_Xdatas[i]]]
            y = [e[1] for e in Xdatas[i_Xdatas[i]]]
            ax.quiver(
                x,
                y,
                Ydatas[0][:, 0],
                Ydatas[0][:, 1],
                color=color_list[i],
                scale_units=scale_units,
                scale=scale,
                width=width,
                minshaft=2,
                zorder=10,
                label=legend_list[i],
                headwidth=2,
                headlength=4,
            )
            ax.axis("equal")

    elif type_plot == "curve_point":
        for i in range(ndatas):
            ax.plot(
                Xdatas[i_Xdatas[i]],
                Ydatas[i],
                color=color_list[i],
                label=legend_list[i],
                linewidth=linewidth_list[i],
                picker=True,
                pickradius=5,
            )
            ax.plot(
                Xdatas[i_Xdatas[i]],
                Ydatas[i],
                symbol,
                markerfacecolor=color_list[i],
                markeredgecolor=color_list[i],
                markersize=10,
            )
    elif type_plot == "point":
        for i in range(ndatas):
            ax.plot(
                Xdatas[i_Xdatas[i]],
                Ydatas[i],
                symbol,
                label=legend_list[i],
                markerfacecolor=color_list[i],
                markeredgecolor=color_list[i],
                markersize=10,
                mew=linewidth_list[i],
                picker=True,
                pickradius=5,
            )
    elif type_plot == "barStackResultant":
        data = Ydatas[0:-1]

        def get_cumulated_array(data, **kwargs):
            cum = data.clip(**kwargs)
            cum = cumsum(cum, axis=0)
            d = zeros(shape(data))
            d[1:] = cum[:-1]
            return d

        cumulated_data = get_cumulated_array(data, min=0)
        cumulated_data_neg = get_cumulated_array(data, max=0)

        # Re-merge negative and positive data.
        row_mask = data < 0
        cumulated_data[row_mask] = cumulated_data_neg[row_mask]
        data_stack = cumulated_data

        for i in range(len(data)):
            ax.bar(
                x=Xdatas[0],
                height=data[i],
                bottom=data_stack[i],
                color=color_list[i],
                label=legend_list[i],
                width=barwidth,
                picker=True,
            )

        ax.bar(
            x=Xdatas[1],
            height=Ydatas[-1],
            width=barwidth,
            color=["k"],
            label="Resultant",
            picker=True,
        )

        ax.legend()

        if xticks is None:
            plt.xticks([])

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xlim([x_min, x_max])
    ax.set_ylim([y_min, y_max])

    if is_logscale_x:
        ax.set_xscale("log")

    if is_logscale_y:
        ax.set_yscale("log")

    if is_disp_title:
        ax.set_title(title)

    if is_grid:
        ax.grid()

    # if ndatas > 1 and not no_legend:
    if not no_legend:
        ax.legend(prop={"family": font_name, "size": font_size_legend})

    if not is_show_legend:
        ax.get_legend().remove()

    plt.tight_layout()
    for item in (
        [ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()
    ):
        item.set_fontname(font_name)
        item.set_fontsize(font_size_label)
    ax.title.set_fontname(font_name)
    ax.title.set_fontsize(font_size_title)

    if save_path is not None:
        save_path = save_path.replace("\\", "/")
        fig.savefig(save_path)
        plt.close()

    if is_show_fig:
        fig.show()

    if win_title:
        manager = plt.get_current_fig_manager()
        if manager is not None:
            manager.set_window_title(win_title)



# -*- coding: utf-8 -*-
class FTError(Exception):
    """Raised when there is an error within the Fourier Transform methods (Data Object)"""

    pass


class AxisError(FTError):
    """Raised when the axis requested is not available (Data Object)"""

    pass


class UnitError(FTError):
    """Raised when the unit requested is not available or does not match the existing one (Data Object)"""

    pass


class NormError(FTError):
    """Raised when the unit requested is not available or does not match the existing one (Data Object)"""

    pass


axes_dict = {
    "angle": ["wavenumber", "ifft", "rad"],
    "time": ["freqs", "ifft", "s"],
    "xyz": ["rphiz", "pol2cart", "m"],
}

rev_axes_dict = {
    "wavenumber": ["angle", "fft", "dimless"],
    "freqs": ["time", "fft", "Hz"],
    "rphiz": ["xyz", "cart2pol", "m"],
}
